#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import asyncio
import logging
import re
from asyncio.exceptions import CancelledError
from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty
from random import choice
from string import ascii_lowercase
from threading import Thread
from time import sleep, monotonic, time
from typing import List, Dict, Union
from typing import Tuple, Any
from cachetools import TTLCache

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.gateway.constants import CONNECTOR_PARAMETER, RECEIVED_TS_PARAMETER, CONVERTED_TS_PARAMETER, \
    DATA_RETRIEVING_STARTED, REPORT_STRATEGY_PARAMETER, ON_ATTRIBUTE_UPDATE_DEFAULT_TIMEOUT
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_logger import init_logger
from thingsboard_gateway.tb_utility.tb_utility import TBUtility

# Try import asyncua library or install it and import
installation_required = False
required_version = '1.1.5'
force_install = False

from packaging import version
try:
    from asyncua import __version__ as asyncua_version

    if version.parse(asyncua_version) != version.parse(required_version):
        installation_required = True
        force_install = True

except ImportError:
    installation_required = True

if installation_required:
    print("OPC-UA library not found")
    TBUtility.install_package("asyncua", required_version, force_install=force_install)

import asyncua
from asyncua import ua, Node
from asyncua.ua import NodeId, UaStringParsingError
from asyncua.common.ua_utils import value_to_datavalue
from asyncua.ua.uaerrors import BadWriteNotSupported, BadTypeMismatch
from asyncua.crypto.security_policies import SecurityPolicyBasic256Sha256, SecurityPolicyBasic256, \
    SecurityPolicyBasic128Rsa15
from asyncua.ua.uaerrors import UaStatusCodeError, BadNodeIdUnknown, BadConnectionClosed, \
    BadInvalidState, BadSessionClosed, BadAttributeIdInvalid, BadCommunicationError, BadOutOfService, BadNoMatch, \
    BadUnexpectedError, UaStatusCodeErrors, BadWaitingForInitialData, BadSessionIdInvalid, BadSubscriptionIdInvalid
from thingsboard_gateway.connectors.opcua.entities.rpc_request import OpcUaRpcRequest, OpcUaRpcType
from thingsboard_gateway.connectors.opcua.device import Device
from thingsboard_gateway.connectors.opcua.backward_compatibility_adapter import BackwardCompatibilityAdapter

DEFAULT_UPLINK_CONVERTER = 'OpcUaUplinkConverter'

SECURITY_POLICIES = {
    "Basic128Rsa15": SecurityPolicyBasic128Rsa15,
    "Basic256": SecurityPolicyBasic256,
    "Basic256Sha256": SecurityPolicyBasic256Sha256,
}

MESSAGE_SECURITY_MODES = {
    "None": asyncua.ua.MessageSecurityMode.None_,
    "Sign": asyncua.ua.MessageSecurityMode.Sign,
    "SignAndEncrypt": asyncua.ua.MessageSecurityMode.SignAndEncrypt
}

RPC_SET_SPLIT_PATTERNS = ["; ", ";=", "=", " "]


class OpcUaConnector(Connector, Thread):
    def __init__(self, gateway: 'TBGatewayService', config, connector_type):
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        super().__init__()
        self.__scanning_nodes_cache = TTLCache(maxsize=100_000, ttl=3600)
        self._connector_type = connector_type
        self.__gateway: 'TBGatewayService' = gateway
        self.__config = config
        self.__id = self.__config.get('id')
        self.name = self.__config.get("name", 'OPC-UA Connector ' + ''.join(choice(ascii_lowercase) for _ in range(5)))
        self.__last_contact_time = 0

        # check if config is in old format and convert it to new format
        using_old_configuration_format = len(tuple(filter(lambda node_config: not node_config.get('deviceInfo'),
                                                          self.__config.get('server', {}).get('mapping', []))))
        self.__enable_remote_logging = self.__config.get('enableRemoteLogging', False)
        self.__log = init_logger(self.__gateway, self.name, self.__config.get('logLevel', 'INFO'),
                                 enable_remote_logging=self.__enable_remote_logging, is_connector_logger=True)
        self.__converter_log = init_logger(self.__gateway, self.name + '_converter',
                                           self.__config.get('logLevel', 'INFO'),
                                           enable_remote_logging=self.__enable_remote_logging,
                                           is_converter_logger=True, attr_name=self.name)
        self.__replace_loggers()
        if gateway.get_report_strategy_service() is not None:
            report_strategy = self.__config.get('reportStrategy')
            self.__connector_report_strategy_config = gateway.get_report_strategy_service().get_main_report_strategy()
            try:
                if report_strategy is not None:
                    self.__connector_report_strategy_config = ReportStrategyConfig(report_strategy)
            except ValueError as e:
                self.__log.warning(
                    'Error in report strategy configuration: %s, the gateway main strategy will be used.',
                    e)
        if using_old_configuration_format:
            backward_compatibility_adapter = BackwardCompatibilityAdapter(self.__config, self.__log)
            self.__config = backward_compatibility_adapter.convert()

        self.__server_conf = self.__config.get('server', {})
        self.__server_limits = {}
        self.__max_nodes_per_read = 100
        self.__max_nodes_per_subscribe = 100

        if using_old_configuration_format:
            self.__log.warning('Connector configuration has been updated to the new format.')

        if "opc.tcp" not in self.__server_conf.get("url"):
            self.__opcua_url = "opc.tcp://" + self.__server_conf.get("url")
        else:
            self.__opcua_url = self.__server_conf.get("url")

        self.__enable_subscriptions = self.__server_conf.get('enableSubscriptions', True)
        self.__sub_check_period_in_millis = max(self.__server_conf.get("subCheckPeriodInMillis", 100), 100)
        # Batch size for data change subscription, the gateway will process this amount of data, received from subscriptions, or less in one iteration
        self.__sub_data_max_batch_size = self.__server_conf.get("subDataMaxBatchSize", 1000)
        self.__sub_data_min_batch_creation_time = max(self.__server_conf.get("subDataMinBatchCreationTimeMs", 200),
                                                      100) / 1000
        self.__subscription_batch_size = self.__server_conf.get('subscriptionProcessBatchSize', 2000)

        self.__reconnect_retries_count = self.__server_conf.get('reconnectRetriesCount', 8)
        self.__reconnect_backoff_initial_delay = self.__server_conf.get('reconnectBackoffInitialDelay', 1)
        self.__reconnect_backoff_factor = self.__server_conf.get('reconnectBackoffFactor', 2)

        self.__show_map = self.__server_conf.get('showMap', False)

        self.__sub_data_to_convert = Queue(-1)
        self.__data_to_convert = Queue(-1)

        try:
            self.__loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.__loop)
        except RuntimeError:
            self.__loop = asyncio.get_event_loop()

        self.__client: Union[asyncua.Client, None] = None

        self.__connected = False
        self.__stopped = False
        self.daemon = True

        self.__thread_pool_executor = ThreadPoolExecutor(max_workers=8)
        self.__thread_pool_executor_processor_thread = Thread(name='Thread Pool Executor Processor',
                                                              target=self.__thread_pool_executor_processor,
                                                              daemon=True)
        self.__thread_pool_executor_processor_thread.start()

        self.__device_nodes: List[Device] = []
        self.__nodes_config_cache: Dict[NodeId, List[Device]] = {}
        self.__next_poll = 0
        self.__next_scan = 0
        self.__client_recreation_required = True

        self.__log.info("OPC-UA Connector has been initialized")

    def __replace_loggers(self):
        for logger_name in logging.root.manager.loggerDict.keys():
            if 'asyncua' in logger_name:
                init_logger(self.__gateway,
                            logger_name,
                            'ERROR',
                            enable_remote_logging=self.__enable_remote_logging,
                            is_connector_logger=True,
                            attr_name=self.name)

    def get_device_shared_attributes_keys(self, device_name):
        device = self.__get_device_by_name(device_name)
        if device is not None:
            return device.shared_attributes_keys
        return []

    def open(self):
        self.__stopped = False
        self.start()
        self.__log.info("Starting OPC-UA Connector (Async IO)")

    def get_type(self):
        return self._connector_type

    def close(self):
        self.__stopped = True
        self.__connected = False
        self.__log.info("Stopping OPC-UA Connector")

        self.__delete_devices_from_platform()
        asyncio.run_coroutine_threadsafe(self.__disconnect(), self.__loop)
        asyncio.run_coroutine_threadsafe(self.__cancel_all_tasks(), self.__loop)

        start_time = monotonic()

        while self.is_alive():
            if monotonic() - start_time > 10:
                self.__log.error("Failed to stop connector %s", self.get_name())
                break
            sleep(.1)

        self.__log.info('%s has been stopped.', self.get_name())
        self.__log.stop()

    async def __cancel_all_tasks(self):
        await asyncio.sleep(5)
        for task in asyncio.all_tasks(self.__loop):
            task.cancel()

    async def __disconnect(self):
        try:
            if self.__connected:
                await self.__unsubscribe_from_nodes()
            await self.__client.disconnect()
            self.__log.info('%s has been disconnected from OPC-UA Server.', self.get_name())
        except Exception as e:
            self.__log.warning('%s could not be disconnected from OPC-UA Server: %s', self.name, e)

    async def __unsubscribe_from_node(self, device: Device, node):
        node.pop('valid', None)
        if node.get('node') is not None and self.__enable_subscriptions:
            subscription_id = device.nodes_data_change_subscriptions.get(node['node'].nodeid, {}).get('subscription')
            if subscription_id is not None:
                try:
                    await device.subscription.unsubscribe(subscription_id)
                except Exception as e:
                    self.__log.exception('Error unsubscribing from on data change: %s', e)
                device.nodes_data_change_subscriptions[node['node'].nodeid]['subscription'] = None

    async def __unsubscribe_from_nodes(self, device_name=None):
        for device in self.__device_nodes:
            if device_name is None or device.name == device_name:
                for section in ('attributes', 'timeseries'):
                    for node in device.values.get(section, []):
                        await self.__unsubscribe_from_node(device, node)

            if (device_name is None and device.subscription is not None
                    and self.__client.uaclient.protocol is not None
                    and self.__client.uaclient.protocol.state == 'open'):
                try:
                    await device.subscription.delete()
                except AttributeError:
                    self.__log.debug('Subscription already deleted')
                except BadSubscriptionIdInvalid:
                    self.__log.debug('Subscription already deleted')
                except ConnectionError:
                    self.__log.info("Client is not connected, cannot delete subscription.")
                except Exception as e:
                    self.__log.exception('Error deleting subscription: %s', e)
                device.subscription = None
        self.__nodes_config_cache.clear()

    def get_id(self):
        return self.__id

    def get_name(self):
        return self.name

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.__stopped

    def get_config(self):
        return self.__config

    def run(self):

        if self.__enable_subscriptions:
            sub_data_convert_thread = Thread(name='Sub Data Convert Thread', target=self.__convert_sub_data,
                                             daemon=True)
            sub_data_convert_thread.start()

        try:
            self.__loop.run_until_complete(self.start_client())
        except asyncio.CancelledError:
            try:
                self.__loop.run_until_complete(self.disconnect_if_connected())
            except Exception:
                pass
        except Exception as e:
            self.__log.exception("Error in main loop: %s", e)

    async def start_client(self):
        sleep_for_subscription_work_model = max(self.__sub_check_period_in_millis / 1000, .02)
        while not self.__stopped:
            try:
                reconnect_required = True
                if self.__client_recreation_required:
                    if (self.__client is not None
                            and self.__client.uaclient.protocol
                            and self.__client.uaclient.protocol.state == 'open'):
                        await self.disconnect_if_connected()
                    self.__client = asyncua.Client(url=self.__opcua_url,
                                                   timeout=self.__server_conf.get('timeoutInMillis', 4000) / 1000)
                    self.__client._monitor_server_loop = self._monitor_server_loop
                    self.__client._renew_channel_loop = self._renew_channel_loop
                    self.__client.session_timeout = self.__server_conf.get('sessionTimeoutInMillis', 120000)
                    if self.__server_conf["identity"].get("type") == "cert.PEM":
                        await self.__set_auth_settings_by_cert()
                    if self.__server_conf["identity"].get("username"):
                        self.__set_auth_settings_by_username()
                    if self.__stopped:
                        break
                    if self.__enable_subscriptions and self.__device_nodes:
                        self.__log.debug(
                            "Subscriptions are enabled, client will reconnect, unsubscribe old subscriptions and subscribe to new nodes.")
                        await self.retry_connect_with_backoff(self.__reconnect_retries_count,
                                                              self.__reconnect_backoff_initial_delay,
                                                              self.__reconnect_backoff_factor)
                        self.__last_contact_time = monotonic()
                        if not (self.__client.uaclient.protocol and self.__client.uaclient.protocol.state == 'open'):
                            self.__log.error("Failed to connect to server, retrying...")
                            await self.disconnect_if_connected()
                            continue
                        await self.__unsubscribe_from_nodes()
                        reconnect_required = False

                    self.__nodes_config_cache = {}
                    self.__delete_devices_from_platform()
                    self.__device_nodes = []
                    self.__next_scan = 0
                    self.__next_poll = 0
                    self.__client_recreation_required = False

                if reconnect_required:
                    await self.retry_connect_with_backoff(self.__reconnect_retries_count,
                                                          self.__reconnect_backoff_initial_delay,
                                                          self.__reconnect_backoff_factor)
                    self.__last_contact_time = monotonic()

                if not (self.__client.uaclient.protocol and self.__client.uaclient.protocol.state == 'open'):
                    self.__log.error("Failed to connect to server, retrying...")
                    await self.disconnect_if_connected()
                    continue
                self.__log.info("Connected to OPC-UA Server: %s", self.__opcua_url)
                self.__connected = True

                try:
                    await self.__client.load_data_type_definitions()
                except Exception as e:
                    self.__log.error("Error on loading type definitions:\n %s", e)

                try:
                    await self.__fetch_server_limitations()
                except Exception as e:
                    self.__log.error("Error on fetching server limitations:\n %s", e)

                poll_period = int(self.__server_conf.get('pollPeriodInMillis', 5000) / 1000)
                scan_period = int(self.__server_conf.get('scanPeriodInMillis', 3600000) / 1000)

                if self.__enable_subscriptions:
                    await self.__scan_device_nodes()
                    self.__next_scan = monotonic() + scan_period
                    # await self.__poll_nodes()

                while not self.__stopped:
                    if self.__client_recreation_required:
                        await self.disconnect_if_connected()
                        break
                    if monotonic() >= self.__next_scan:
                        self.__next_scan = monotonic() + scan_period
                        await self.__scan_device_nodes()

                    if not self.__enable_subscriptions and monotonic() >= self.__next_poll:
                        self.__next_poll = monotonic() + poll_period
                        await self.__poll_nodes()

                    current_time = monotonic()
                    time_to_sleep = min(self.__next_poll - current_time, self.__next_scan - current_time) \
                        if not self.__enable_subscriptions else sleep_for_subscription_work_model
                    if time_to_sleep > 0:
                        await asyncio.sleep(time_to_sleep)
                    if not self.__connected:
                        if self.__stopped:
                            self.__log.debug("Connection closed due to stopping.")
                        else:
                            self.__log.debug("Detected connection lost, OPC-UA client will reconnect to server.")
                            await self.disconnect_if_connected()
                        break

            except (ConnectionError, BadSessionClosed, BadSessionIdInvalid):
                self.__log.warning('Connection lost for %s, will try to reconnect...', self.get_name())
                await self.disconnect_if_connected()
                self.__client_recreation_required = True
            except asyncio.exceptions.TimeoutError:
                self.__log.warning('Failed to connect %s', self.get_name())
                self.__client_recreation_required = True
                await self.disconnect_if_connected()
                await asyncio.sleep(5)
            except asyncio.CancelledError as e:
                if self.__stopped:
                    self.__log.debug('Task was cancelled due to connector stop: %s', e.__str__())
                else:
                    self.__log.exception('Task was cancelled: %s', e.__str__())
            except UaStatusCodeError as e:
                self.__log.error('Error in main loop, trying to reconnect: %s', exc_info=e)
                await self.disconnect_if_connected()
            except Exception as e:
                self.__log.exception("Error in main loop: %s", e)
            finally:
                if self.__stopped:
                    try:
                        await self.disconnect_if_connected()
                    except Exception:
                        pass
                    self.__connected = False
                await asyncio.sleep(.5)

    async def disconnect_if_connected(self):
        if self.__connected:
            try:
                if self.__enable_subscriptions:
                    self.__client_recreation_required = True
                await self.__client.disconnect()
            except Exception:
                pass  # ignore

    def __delete_devices_from_platform(self):
        for device in self.__device_nodes:
            self.__gateway.del_device(device.name)

    async def retry_connect_with_backoff(self, max_retries=8, initial_delay=1, backoff_factor=2):
        last_contact_delta = monotonic() - self.__last_contact_time
        if last_contact_delta < self.__client.session_timeout / 1000 and self.__last_contact_time > 0:
            time_to_wait = self.__client.session_timeout / 1000 - last_contact_delta
            time_to_wait = min(self.__server_conf.get('sessionTimeoutInMillis', 120000) / 1000 * 0.8, time_to_wait)
            self.__log.info('Last contact was %.2f seconds ago, next connection try in %.2f seconds...',
                            last_contact_delta, time_to_wait)
            await asyncio.sleep(time_to_wait)
        delay = initial_delay
        for attempt in range(max_retries):
            if self.__stopped:
                return None
            try:
                return await self.__client.connect()
            except Exception as e:
                base_time = self.__client.session_timeout / 1000 if (
                        last_contact_delta > 0 and last_contact_delta < self.__client.session_timeout / 1000) else 0
                time_to_wait = base_time + delay
                time_to_wait = min(self.__server_conf.get('sessionTimeoutInMillis', 120000) / 1000 * 0.8, time_to_wait)
                self.__log.error('Encountered error: %r. Next connection try in %i second(s)...', e, time_to_wait)
                await asyncio.sleep(time_to_wait)
                delay *= backoff_factor
        self.__log.error('Max retries reached. Connection failed.')
        return None

    async def __set_auth_settings_by_cert(self):
        try:
            ca_cert = self.__server_conf["identity"].get("caCert")
            private_key = self.__server_conf["identity"].get("privateKey")
            cert = self.__server_conf["identity"].get("cert")
            policy = self.__server_conf["security"]
            mode = self.__server_conf["identity"].get("mode", "SignAndEncrypt")

            if cert is None or private_key is None:
                self.__log.exception("Error in ssl configuration - cert or privateKey parameter not found")
                raise RuntimeError("Error in ssl configuration - cert or privateKey parameter not found")

            await self.__client.set_security(
                SECURITY_POLICIES[policy],
                certificate=cert,
                private_key=private_key,
                server_certificate=ca_cert,
                mode=MESSAGE_SECURITY_MODES[mode]
            )
        except Exception as e:
            self.__log.exception(e)

    def __set_auth_settings_by_username(self):
        self.__client.set_user(self.__server_conf["identity"].get("username"))
        if self.__server_conf["identity"].get("password"):
            self.__client.set_password(self.__server_conf["identity"].get("password"))

    async def _monitor_server_loop(self):
        """
        Checks if the server is alive
        """
        timeout = min(self.__client.session_timeout / 1000 / 2, self.__client._watchdog_intervall)
        try:
            while not self.__client._closing and not self.__stopped:
                await asyncio.sleep(timeout)
                _ = await self.__client.nodes.server_state.read_value()
                self.__last_contact_time = monotonic()
        except ConnectionError as e:
            await self.__client._lost_connection(e)
            await self.__client.uaclient.inform_subscriptions(ua.StatusCode(ua.StatusCodes.BadShutdown))
            self.__connected = False
        except Exception as e:
            await self.__client._lost_connection(e)
            await self.__client.uaclient.inform_subscriptions(ua.StatusCode(ua.StatusCodes.BadShutdown))
            self.__connected = False
        finally:
            if not self.__stopped:
                self.__client_recreation_required = True

    async def _renew_channel_loop(self):
        """
        Renew the SecureChannel before the SecureChannelTimeout will happen.
        In theory, we could do that only if no session activity,
        but it does not cost much.
        """
        try:
            duration = self.__client.secure_channel_timeout * 0.75 / 1000
            while not self.__client._closing and not self.__stopped:
                await asyncio.sleep(duration)
                await self.__client.open_secure_channel(renew=True)
                val = await self.__client.nodes.server_state.read_value()
        except CancelledError:
            self.__log.debug("Renew channel loop task cancelled")
        except ConnectionError as e:
            self.__log.error("Connection error in renew_channel loop %s", e)
        except TimeoutError:
            self.__log.error("Timeout error in renew_channel loop")
        except Exception as e:
            self.__log.exception("Error in renew_channel loop %s", e)
        finally:
            if not self.__stopped:
                self.__client_recreation_required = True

    def __load_converter(self, device):
        converter_class_name = device.get('converter')
        if not converter_class_name:
            self.__log.debug("No custom converter found for device %s, using default converter", device.get('name'))
            converter_class_name = DEFAULT_UPLINK_CONVERTER
        module = TBModuleLoader.import_module(self._connector_type, converter_class_name)

        if module:
            self.__log.debug('Converter %s for device %s - found!', converter_class_name, self.name)
            return module

        self.__log.error("Cannot find converter for %s device", self.name)
        return None

    @staticmethod
    def is_regex_pattern(pattern):
        return not re.fullmatch(pattern, pattern)

    async def __find_nodes(self, node_list_to_search, current_parent_node, nodes, path="Root"):
        assert len(node_list_to_search) > 0
        final = []

        target_node_path = None
        if len(node_list_to_search) == 1:
            node_paths = [node['path'] if isinstance(node, dict) else node for node in nodes]
            target_node_path = '.'.join(node_path.split(':')[-1] for node_path in node_paths) + '.' + \
                               node_list_to_search[0]
            if target_node_path in self.__scanning_nodes_cache:
                if self.__show_map:
                    self.__log.debug('Found node in cache: %s', node_list_to_search[0])
                final.append(self.__scanning_nodes_cache[target_node_path])
                return final

        children = await current_parent_node.get_children()
        children_nodes_count = len(children)
        counter = 0
        if self.__show_map:
            self.__log.debug('Found %s children for %s', children_nodes_count, current_parent_node)
        for node in children:
            counter += 1
            child_node = await node.read_browse_name()
            if len(node_list_to_search) == 1:
                node_paths = [node['path'] if isinstance(node, dict) else node for node in nodes]
                current_node_path = '.'.join(
                    node_path.split(':')[-1] for node_path in node_paths) + '.' + child_node.Name
                if current_node_path not in self.__scanning_nodes_cache:
                    self.__scanning_nodes_cache[current_node_path] = [*nodes, {
                        'path': f'{child_node.NamespaceIndex}:{child_node.Name}', 'node': node}]
            if self.__show_map and path:
                if children_nodes_count < 1000 or counter % 1000 == 0:
                    self.__log.info('Checking path: %s', path + '.' + f'{child_node.Name}')

            if re.fullmatch(re.escape(node_list_to_search[0]), child_node.Name) or node_list_to_search[0].split(':')[-1] == child_node.Name:
                if self.__show_map:
                    self.__log.info('Found node: %s', child_node.Name)
                new_nodes = [*nodes, {'path': f'{child_node.NamespaceIndex}:{child_node.Name}', 'node': node}]
                if len(node_list_to_search) == 1:
                    final.append(new_nodes)
                    if self.__show_map:
                        self.__log.debug('Found node: %s', child_node.Name)
                    return final
                else:
                    final.extend(
                        await self.__find_nodes(node_list_to_search[1:], current_parent_node=node, nodes=new_nodes,
                                                path=path + '.' + f'{child_node.Name}'))

        return final

    async def find_nodes(self, node_pattern, current_parent_node=None, nodes=None):
        if nodes is None:
            nodes = []
        node_list = node_pattern.split('\\.')

        if current_parent_node is None:
            current_parent_node = self.__client.nodes.root

            if len(node_list) > 0 and node_list[0].lower() == 'root':
                node_list = node_list[1:]

        return await self.__find_nodes(node_list, current_parent_node, nodes, 'Root')

    async def find_node_name_space_index(self, path):
        if isinstance(path, str):
            path = path.split('\\.')

        # find unresolved nodes
        u_node_count = len(tuple(filter(lambda u_node: len(u_node.split(':')) < 2, path)))

        resolved = path[:-u_node_count]
        resolved_level = len(path) - u_node_count
        if resolved_level < 1:
            path = path[1:]
            parent_node = self.__client.get_root_node()
        else:
            parent_node = await self.__client.nodes.root.get_child(resolved)

        unresolved = path[resolved_level:]
        return await self.__find_nodes(unresolved, current_parent_node=parent_node, nodes=resolved)

    async def _get_device_info_by_pattern(self, pattern, get_first=False, parent_node=None):
        """
        Method used to retrieve device name/profile by pattern.
        Device name/profile can consist of path expressions or NodeId expressions but not both.
        It will search for device name in the following order:
        1. By path expressions (e.g. "My Device ${Root\\.Objects\\.Device\\.serialNumber}}")
        2. By NodeId expressions (e.g. "My Device ${ns=2;i=1003} - ${ns=2;i=1008}")
        3. If no matches found, returns the pattern as is.
        """

        search_results = await self.__get_device_expression_value_by_path(pattern, parent_node=parent_node)
        if len(search_results) > 0:
            return search_results[0] if get_first else search_results

        search_results = await self.__get_device_expression_value_by_identifier(pattern)
        if len(search_results) > 0:
            return search_results[0] if get_first else search_results

        return [pattern]

    async def __get_device_expression_value_by_path(self, pattern: str, parent_node=None) -> List[str]:
        results = []

        search_result = re.search(r"\${([A-Za-z.:_\-\\\d\[\]]+)}", pattern)
        if search_result:
            try:
                group = search_result.group(0)
                node_path = search_result.group(1)
            except IndexError:
                self.__log.error('Invalid pattern: %s', pattern)
                return results

            nodes = await self.find_nodes(node_path, current_parent_node=parent_node)
            self.__log.debug('Found device name nodes: %s', nodes)

            for node in nodes:
                try:
                    var = node[-1]['node']
                    value = await var.read_value()
                    results.append(pattern.replace(group, str(value)))
                except Exception as e:
                    self.__log.exception(e)
                    continue

        return results

    async def __get_device_expression_value_by_identifier(self, pattern: str) -> List[str]:
        result = pattern
        node_ids_search_result = re.findall(r"\${(ns=\d+;[isgb]=[^}]+)}", pattern)
        if node_ids_search_result:
            for group in node_ids_search_result:
                try:
                    node_id = NodeId.from_string(group)
                    value = await self.__client.get_node(node_id).read_value()
                    result = result.replace(f'${{{group}}}', str(value))
                except Exception as e:
                    self.__log.error('Invalid NodeId format %s: %s', group, e)

        return [result]

    def __convert_sub_data(self):
        device_converted_data_map = {}
        sleep_period_after_empty_batch = max(self.__sub_check_period_in_millis / 1000, .02)

        while not self.__stopped:
            batch = []
            batch_start_forming_time = time()
            while (not self.__sub_data_to_convert.empty()
                   and len(batch) < self.__sub_data_max_batch_size
                   and time() - batch_start_forming_time < self.__sub_data_min_batch_creation_time):
                try:
                    batch.append(self.__sub_data_to_convert.get_nowait())
                    if (time() - batch_start_forming_time) >= self.__sub_data_min_batch_creation_time:
                        break
                except Empty:
                    break

            if not batch and self.__sub_data_to_convert.empty():
                sleep(sleep_period_after_empty_batch)
                continue
            if self.__log.isEnabledFor(5):
                self.__log.debug('Batch created in %i milliseconds', (time() - batch_start_forming_time) * 1000)
                self.__log.debug('Batch size: %s', len(batch))
                self.__log.debug('Data left in queue: %s', self.__sub_data_to_convert.qsize())

            for sub_node, data, received_ts in batch:
                node_configs = self.__nodes_config_cache.get(sub_node.nodeid, [])
                for device in node_configs:
                    try:
                        if sub_node.nodeid not in device.nodes_data_change_subscriptions:
                            continue
                        if device not in device_converted_data_map:
                            device_converted_data_map[device] = ConvertedData(device_name=device.name,
                                                                              device_type=device.device_profile)

                        nodes_configs = device.nodes_data_change_subscriptions[sub_node.nodeid]['nodes_configs']
                        nodes_values = [data.monitored_item.Value for _ in range(len(nodes_configs))]

                        converted_data = device.converter_for_sub.convert(nodes_configs, nodes_values)

                        if converted_data:
                            converted_data.add_to_metadata({
                                CONNECTOR_PARAMETER: self.get_name(),
                                RECEIVED_TS_PARAMETER: received_ts,
                                CONVERTED_TS_PARAMETER: int(time() * 1000)
                            })

                            for node_config in nodes_configs:
                                if node_config['section'] == 'attributes':
                                    device_converted_data_map[device].add_to_attributes(converted_data.attributes)
                                else:
                                    device_converted_data_map[device].add_to_telemetry(converted_data.telemetry)
                    except Exception as e:
                        self.__log.exception("Error converting data: %s", e)

            for device, converted_data in device_converted_data_map.items():
                self.__gateway.send_to_storage(self.get_name(), self.get_id(), converted_data)
                self.__log.debug('Converted data from %r notifications from server for device %s, using converter: %r',
                                 len(batch), device.name, device.converter_for_sub.__class__.__name__)

            device_converted_data_map.clear()

    async def __scan_device_nodes(self):
        await self._create_new_devices()
        await self._load_devices_nodes()

    async def __get_device_base_nodes(self, device_node_pattern):
        try:
            if self.__is_node_identifier(device_node_pattern):
                node_id = NodeId.from_string(device_node_pattern)
                node = self.__client.get_node(node_id)
                node_browse_name = await node.read_browse_name()
                return [[{'path': f'{node_browse_name.NamespaceIndex}:{node_browse_name.Name}', 'node': node}]]
            elif self.__is_absolute_path(device_node_pattern):
                nodes = await self.find_nodes(device_node_pattern)
                return nodes
            else:
                return []
        except Exception as e:
            self.__log.error('Error finding device base nodes by pattern %s: %s', device_node_pattern, e)
            return []

    async def _create_new_devices(self):
        self.__log.debug('Scanning for new devices from configuration...')
        existing_devices = list(map(lambda dev: dev.name, self.__device_nodes))

        scanned_devices = []
        for device_config in self.__config.get('mapping', []):
            nodes = await self.__get_device_base_nodes(device_config['deviceNodePattern'])
            if not nodes:
                self.__log.error(
                    "No nodes found for such device name '%s' "
                    "Skipping device creation.",
                    device_config.get('deviceInfo', {}).get('deviceNameExpression', "")
                )
                continue

            self.__log.debug('Found devices: %s', nodes)

            for node in nodes:
                device_name_expression = device_config.get('deviceInfo', {}).get('deviceNameExpression')

                is_abs_path = self.__is_absolute_path(device_name_expression)
                device_names = await self._get_device_info_by_pattern(device_name_expression,
                                                                      parent_node=node[-1]['node'] if not is_abs_path else None)

                for device_name in device_names:
                    scanned_devices.append(device_name)

                    if device_name not in existing_devices:
                        converter = self.__load_converter(device_config)

                        device_profile_expression = device_config.get('deviceInfo', {}).get('deviceProfileExpression',
                                                                                            'default')
                        is_abs_path = self.__is_absolute_path(device_profile_expression)
                        device_profile = await self._get_device_info_by_pattern(device_profile_expression,
                                                                                get_first=True,
                                                                                parent_node=node[-1]['node'] if not is_abs_path else None)

                        device_config = {**device_config, 'device_name': device_name, 'device_type': device_profile}
                        device_path = [node_path_node_object['path'] for node_path_node_object in node]
                        self.__device_nodes.append(
                            Device(path=device_path, name=device_name, device_profile=device_profile,
                                   config=device_config,
                                   converter=converter(
                                       device_config, self.__converter_log),
                                   converter_for_sub=converter(device_config,
                                                               self.__converter_log) if self.__enable_subscriptions else None,
                                   device_node=node[-1]['node'],
                                   logger=self.__log))
                        self.__log.info('Added device node: %s', device_name)
        self.__log.debug('Device nodes: %s', self.__device_nodes)

    async def _load_devices_nodes(self):
        for device in self.__device_nodes:
            device.nodes = []
            for section in ('attributes', 'timeseries'):
                self.__log.info('Loading nodes for device: %s, section: %s, nodes count: %s', device.name, section,
                                len(device.values.get(section, [])))
                for node in device.values.get(section, []):
                    try:
                        path = node.get('qualified_path', node['path'])
                        if self.__is_node_identifier(path):
                            found_node = self.__client.get_node(path)
                        else:
                            if not isinstance(path, Node):
                                if len(path[-1].split(':')) != 2:
                                    qualified_path = await self.find_node_name_space_index(path)
                                    if len(qualified_path) == 0:
                                        if node.get('valid', True):
                                            self.__log.warning('Node not found; device: %s, key: %s, path: %s',
                                                               device.name,
                                                               node['key'], node['path'])
                                            if node.get('node') is not None and self.__enable_subscriptions:
                                                await self.__unsubscribe_from_node(device, node)
                                        continue
                                    elif len(qualified_path) > 1:
                                        self.__log.warning(
                                            'Multiple matching nodes found; device: %s, key: %s, path: %s; %s',
                                            device.name,
                                            node['key'], node['path'], qualified_path)
                                    node['qualified_path'] = qualified_path[0][-1]['node']
                                    path = qualified_path[0][-1]['node']

                            if isinstance(path, Node):
                                found_node = path
                            else:
                                found_node = await self.__client.nodes.root.get_child(path)

                        node_report_strategy = node.get(REPORT_STRATEGY_PARAMETER)
                        if self.__gateway.get_report_strategy_service() is not None:
                            if node_report_strategy is not None:
                                try:
                                    node_report_strategy = ReportStrategyConfig(node_report_strategy)
                                except ValueError as e:
                                    self.__log.error(
                                        'Error in report strategy configuration: %s, for key %s the device or connector report strategy will be used.',
                                        e, node['key'])
                                    node_report_strategy = self.__connector_report_strategy_config if device.report_strategy is None else device.report_strategy
                            elif device.report_strategy is not None:
                                node_report_strategy = device.report_strategy

                        node_config = {"node": found_node, "key": node['key'],
                                       "section": section,
                                       'timestampLocation': node.get('timestampLocation', 'gateway')}
                        if self.__gateway.get_report_strategy_service() is not None and node_report_strategy is not None:
                            node_config[REPORT_STRATEGY_PARAMETER] = node_report_strategy
                            node_report_strategy = None  # Cleaning for next iteration

                        device.nodes.append(node_config)

                        subscription_exists = (device.subscription is not None
                                               and found_node.nodeid in device.nodes_data_change_subscriptions
                                               and device.nodes_data_change_subscriptions[found_node.nodeid][
                                                   'subscription'] is not None)

                        if node.get('valid') is None or (node.get('valid') and not self.__enable_subscriptions):
                            if self.__enable_subscriptions and not subscription_exists and not self.__stopped:
                                if found_node.nodeid not in device.nodes_data_change_subscriptions:
                                    if found_node.nodeid not in self.__nodes_config_cache:
                                        self.__nodes_config_cache[found_node.nodeid] = []
                                    self.__nodes_config_cache[found_node.nodeid].append(device)

                                    device_node_config = {
                                        'subscription': None,
                                        'node': found_node,
                                        'nodes_configs': []
                                    }
                                    device.nodes_data_change_subscriptions[found_node.nodeid] = device_node_config

                                device.nodes_data_change_subscriptions[found_node.nodeid]['nodes_configs'].append(
                                    node_config)

                                if device.subscription is None:
                                    device.subscription = await self.__client.create_subscription(
                                        self.__sub_check_period_in_millis, SubHandler(
                                            self.__sub_data_to_convert, self.__log, self.status_change_callback))

                                node['id'] = found_node.nodeid.to_string()

                            node['valid'] = True
                    except ConnectionError as e:
                        raise e
                    except (BadNodeIdUnknown, BadConnectionClosed, BadInvalidState, BadAttributeIdInvalid,
                            BadCommunicationError, BadOutOfService, BadNoMatch, BadUnexpectedError,
                            UaStatusCodeErrors,
                            BadWaitingForInitialData):
                        if node.get('valid', True):
                            self.__log.warning('Node not found (2); device: %s, key: %s, path: %s',
                                               device.name,
                                               node['key'], node['path'])
                            await self.__unsubscribe_from_node(device, node)
                    except UaStatusCodeError as uae:
                        if node.get('valid', True):
                            self.__log.exception('Node status code error: %s', uae)
                            await self.__unsubscribe_from_node(device, node)
                    except Exception as e:
                        if node.get('valid', True):
                            self.__log.exception(e)
                            await self.__unsubscribe_from_node(device, node)
            try:
                if self.__enable_subscriptions and device.subscription is not None:
                    to_subscribe = zip(*list(map(lambda n: (n[1]['node'], n),
                                                 filter(lambda n: n[1]['subscription'] is None,
                                                        device.nodes_data_change_subscriptions.items()))))
                    nodes_to_subscribe = []
                    conf = []
                    if to_subscribe:
                        try:
                            nodes_to_subscribe, conf = to_subscribe
                        except ValueError:
                            pass

                    if nodes_to_subscribe:
                        nodes_data_change_subscriptions = await self._subscribe_for_node_updates_in_batches(
                            device, nodes_to_subscribe,
                            self.__max_nodes_per_subscribe or self.__subscription_batch_size)
                        subs = []
                        for subs_batch in nodes_data_change_subscriptions:
                            subs.extend(subs_batch)
                        for node, conf, subscription_id in zip(nodes_to_subscribe, conf, subs):
                            device.nodes_data_change_subscriptions[node.nodeid]['subscription'] = subscription_id
                        self.__log.info('Subscribed to %i nodes for device %s', len(nodes_to_subscribe), device.name)
                    else:
                        self.__log.debug('No new nodes to subscribe for device %s', device.name)
                else:
                    self.__log.debug('Subscriptions are disabled for device %s or device subscription is None',
                                     device.name)
            except Exception as e:
                self.__log.exception("Error loading nodes: %s", e)
                raise e

    async def _subscribe_for_node_updates_in_batches(self, device, nodes, batch_size=500) -> List:
        total_nodes = len(nodes)
        total_successfully_subscribed = 0
        total_failed_to_subscribe = 0
        result = []
        for i in range(0, total_nodes, batch_size):
            batch = nodes[i:i + batch_size]
            batch_len = len(batch)
            try:
                self.__log.info("Subscribing to batch %i with %i nodes.", i // batch_size + 1, batch_len)
                subscription_result = await device.subscription.subscribe_data_change(batch)
                bad_results = list(filter(lambda r: not isinstance(r, int), subscription_result))
                if bad_results:
                    reasons = [r.doc for r in bad_results]
                    self.__log.error(
                        "Failed subscribing to nodes, server returned the following reasons: %r, nodes count with these problems - %r",
                        set(reasons), len(reasons))
                    self.__log.trace("Failed nodes: %r", bad_results)
                result.append(subscription_result)
                successfully_processed = batch_len - len(bad_results)
                total_successfully_subscribed += successfully_processed
                total_failed_to_subscribe += len(bad_results)
                if successfully_processed > 0:
                    self.__log.info("Succesfully subscribed to batch number %i with %i nodes.", i // batch_size + 1,
                                    successfully_processed)
                else:
                    self.__log.warning("Failed to subscribe to batch number %i with %i nodes.", i // batch_size + 1,
                                       batch_len)
            except Exception as e:
                self.__log.warning("Error subscribing to batch %i with %i : %r", i // batch_size + 1, batch_len, e)
                # self.__log.error("%r", batch) # Uncomment to see the nodes that failed to subscribe
                self.__log.exception("Error subscribing to nodes: ", exc_info=e)
                break
        self.__log.info("Expected subscription to %i nodes.", total_nodes)
        self.__log.info("Successfully subscribed to %i nodes.", total_successfully_subscribed)
        self.__log.info("Failed to subscribe to %i nodes.", total_failed_to_subscribe)
        return result

    def status_change_callback(self, status):
        self.__log.debug('Received status change event: %s', status.Status.doc)
        if status.Status.is_bad():
            self.__client_recreation_required = True

    async def __poll_nodes(self):
        data_retrieving_started = int(time() * 1000)
        all_nodes = [node_config['node'] for device in self.__device_nodes for node_config in device.nodes]

        if len(all_nodes) > 0:
            received_ts = int(time() * 1000)
            for i in range(0, len(all_nodes), self.__max_nodes_per_read):
                batch = all_nodes[i:i + self.__max_nodes_per_read]
                try:
                    values = await self.__client.read_attributes(batch)
                    self.__data_to_convert.put((values, received_ts, data_retrieving_started))
                except Exception as e:
                    self.__log.warning("Failed to read batch from %i to %i: %s", i, i + self.__max_nodes_per_read, e)

        else:
            self.__log.info('No nodes to poll')

    def __thread_pool_executor_processor(self):
        pack = 10
        futures = []
        while not self.__stopped:
            try:
                for future in futures:
                    if future.done():
                        futures.remove(future)
                    else:
                        sleep(.02)
                    if self.__stopped:
                        for future in futures:
                            future.cancel()
                        break
            except Exception as e:
                self.__log.exception("Error in thread pool executor: %s", e)

            try:
                values, received_ts, data_retrieving_started = self.__data_to_convert.get_nowait()
                futures.append(self.__thread_pool_executor.submit(self.__convert_retrieved_data, values, received_ts,
                                                                  data_retrieving_started))
                if len(futures) >= pack:
                    continue
            except Empty:
                sleep(.02)

    def __convert_retrieved_data(self, values, received_ts, data_retrieving_started):
        try:
            converted_nodes_count = 0
            for device in self.__device_nodes:
                nodes_count = len(device.nodes)
                device_values = values[converted_nodes_count:converted_nodes_count + nodes_count]
                converted_nodes_count += nodes_count
                converted_data: ConvertedData = self.__convert_device_data(device.converter, device.nodes,
                                                                           device_values)
                converted_data.add_to_metadata({
                    CONNECTOR_PARAMETER: self.get_name(),
                    RECEIVED_TS_PARAMETER: received_ts,
                    CONVERTED_TS_PARAMETER: int(time() * 1000),
                    DATA_RETRIEVING_STARTED: data_retrieving_started
                })
                if converted_data:
                    self.__gateway.send_to_storage(self.get_name(), self.get_id(), converted_data)

                    StatisticsService.count_connector_message(self.name, stat_parameter_name='connectorMsgsReceived')
                    # TODO: Should these counters be here, or on upper level?
                    StatisticsService.count_connector_bytes(self.name, converted_data,
                                                            stat_parameter_name='connectorBytesReceived')
            self.__log.info('Converted data from %s nodes', converted_nodes_count)
        except Exception as e:
            self.__log.exception("Error converting data: ", exc_info=e)

    @staticmethod
    def __convert_device_data(converter, device_nodes, values):
        return converter.convert(device_nodes, values)

    def __send_data_to_gateway_storage(self, data):
        data.metadata.update({'sendToStorageTs': int(time() * 1000)})

        self.statistics['MessagesReceived'] = self.statistics['MessagesReceived'] + 1
        self.__gateway.send_to_storage(self.get_name(), self.get_id(), data)
        self.statistics['MessagesSent'] = self.statistics['MessagesSent'] + 1
        self.__log.debug('Count data msg to storage: %s', self.statistics['MessagesSent'])

    @staticmethod
    def get_rpc_node_pattern_and_base_path(params, device, logger):
        try:
            if 'Root.' in params:
                current_path = params
                node_pattern = params.replace('.', '\\.')
                return node_pattern, current_path
            names = [v.split(':', 1)[1] for v in device.path]
            node_pattern = "Root\\." + r"\.".join(names) + r'\.' + params
            current_path = 'Root.' + '.'.join(names)
            return node_pattern, current_path
        except Exception as e:
            logger.error("determine_rpc_income_data failed for params=%r: %s",
                         params, e)

    def find_full_node_path(self, params: str, device:Device) -> NodeId | None:
        try:
            node_pattern, current_path = self.get_rpc_node_pattern_and_base_path(params, device, logger=self.__log)
            node_list = node_pattern.split("\\.")[-1:]
            nodes = []
            find_task = self.__find_nodes(node_list, device.device_node, nodes, current_path)
            task = self.__loop.create_task(find_task)
            while not task.done():
                sleep(.1)
            found_nodes = task.result()
            if found_nodes:
                full_path = found_nodes[-1][0]['node'].nodeid
                return full_path
            self.__log.error('Node not found! (%s)', found_nodes)
        except Exception as e:
            self.__log.error("Error during node lookup for %r: %s", params, e)

    def __get_device_by_name(self, device_name: str):
        device_list = tuple(filter(lambda i: i.name == device_name, self.__device_nodes))
        if len(device_list) > 0:
            return device_list[0]

        return None

    def on_attributes_update(self, content: Dict):
        self.__log.debug(content)
        try:
            device = self.__get_device_by_name(content['device'])
            if device is None:
                self.__log.error('Device %s not found for attributes update', content['device'])
                return

            if not device.config.get("attributes_updates"):
                self.__log.error("No attribute mapping found for device %s", device.name)
                return

            for (key, value) in content["data"].items():
                if key not in device.shared_attributes_keys_value_pairs:
                    self.__log.warning("Attribute key %s not found in device attribute section %s", key, device.name)
                    continue

                raw_path = TBUtility.get_value(device.shared_attributes_keys_value_pairs.get(key), get_tag=True)
                node_id = self.__resolve_node_id(raw_path=raw_path, device=device)
                if not isinstance(node_id, NodeId):
                    self.__log.error('Could not find node for device (%s) for key (%s) with path (%s)', device.name,
                                     key, raw_path)

                    continue
                self.__write_node_value(node_id, value, timeout=ON_ATTRIBUTE_UPDATE_DEFAULT_TIMEOUT)
                self.__log.debug("Successfully proccesed attribute update for device %s with key %s", device.name, key)

        except Exception as e:
            self.__log.exception(e)

    def __resolve_node_id(self, raw_path: str, device: Device) -> NodeId | None:
        node_id = (
            NodeId.from_string(raw_path)
            if self.__is_node_identifier(raw_path)
            else self.find_full_node_path(raw_path, device)
        )
        return node_id

    def __write_node_value(self, node_id: NodeId, value, timeout: float) -> bool:
        try:
            write_task = self.__write_value(node_id, value)
            task = self.__loop.create_task(write_task)
            task_completed, result = self.__wait_task_with_timeout(task=task, timeout=timeout,
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.error(
                    "Failed to process request for %s, timeout has been reached", 
                )
                result = {"error": f"Timeout has been reached during write {value}"}

            if err := result.get("error"):
                self.__log.error("Write error on %s: %s", node_id, err)
                return False

            self.__log.debug("Successfully wrote %s to %s", value, node_id)
            return True

        except Exception as exc:
            self.__log.exception("Unexpected error during write: %s", exc)
            self.__log.debug("Unexpected error during write: %s", exc_info=exc)
            return False

    def server_side_rpc_handler(self, content: Dict):
        self.__log.info('Received server side rpc request: %r', content)

        try:
            response = None
            rpc_request = OpcUaRpcRequest(content=content)
            if rpc_request.rpc_type == OpcUaRpcType.CONNECTOR:
                response = self.__process_connector_rpc_request(rpc_request=rpc_request)
            elif rpc_request.rpc_type == OpcUaRpcType.DEVICE:
                response = self.__process_device_rpc_request(rpc_request=rpc_request)

            elif rpc_request.rpc_type == OpcUaRpcType.RESERVED:
                response = self.__process_reserved_rpc_request(rpc_request=rpc_request)
            return response

        except Exception as e:
            self.__log.error('Failed to process server side rpc request: %s', e)
            return {'error': '%r' % e, 'success': False}

    def __process_connector_rpc_request(self, rpc_request: OpcUaRpcRequest):
        self.__log.debug("Received RPC to connector: %r", rpc_request)
        results = []
        for device in self.__device_nodes:
            rpc_request.device = device.name
            try:
                task = self.__loop.create_task(
                    self.__call_method(device.path, rpc_request.rpc_method, rpc_request.arguments))
                task_completed, result = self.__wait_task_with_timeout(task=task, timeout=rpc_request.timeout,
                                                                       poll_interval=0.2)
                if not task_completed:
                    self.__log.error(
                        "Failed to process rpc request for %s, timeout has been reached",
                        device.name,
                    )
                    results.append({"error": f"Timeout rpc has been reached for {device.name}"})
                    continue
                result['device_name'] = device.name
                results.append(result)
                self.__log.debug("RPC with method %s execution result is: %s", rpc_request.rpc_method, result)
            except Exception as e:
                self.__log.exception(e)
                self.__gateway.send_rpc_reply(rpc_request.device, rpc_request.id,
                                              {"result": {"error": str(e)}})
        return results

    @staticmethod
    def __wait_task_with_timeout(task: asyncio.Task, timeout: float, poll_interval: float = 0.2) -> Tuple[bool, Any]:
        start_time = monotonic()
        while not task.done():
            sleep(poll_interval)
            current_time = monotonic()
            if current_time - start_time >= timeout:
                task.cancel()
                return False, None
        return True, task.result()

    def __process_device_rpc_request(self, rpc_request: OpcUaRpcRequest):
        device = self.__get_device_by_name(rpc_request.device_name)
        rpc_section = device.config.get('rpc_methods', [])
        if device is None:
            self.__log.error('Device %s not found for RPC request', rpc_request.device_name)
            self.__gateway.send_rpc_reply(device=rpc_request.device_name,
                                          req_id=rpc_request.id,
                                          content={"result": {"error": 'Device not found'}})
            return

        if not device.is_valid_rpc_method_name(rpc_device_section=rpc_section, rpc_request=rpc_request):
            self.__log.error('Requested rpc method is not found in config %s', rpc_request.device_name)
            self.__gateway.send_rpc_reply(device=rpc_request.device_name,
                                          req_id=rpc_request.id,
                                          content={"result": {"error": 'Requested rpc method is not found in config'}})
            return

        rpc_request.arguments = device.get_device_rpc_arguments(rpc_device_section=rpc_section, rpc_request=rpc_request)
        if isinstance(rpc_request.arguments, dict):
            error_msg = rpc_request.arguments.get('error')
            self.__log.error(f'{error_msg} for device {rpc_request.device_name}')
            self.__gateway.send_rpc_reply(device=rpc_request.device_name,
                                          req_id=rpc_request.id,
                                          content={"result": error_msg})
            return

        try:
            task = self.__loop.create_task(
                self.__call_method(device.path, rpc_request.rpc_method, rpc_request.arguments))
            task_completed, result = self.__wait_task_with_timeout(task=task, timeout=rpc_request.timeout,
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.error(
                    "Failed to process rpc request for %s, timeout has been reached",
                    device.name,
                )
                result = {"error": f"Timeout rpc has been reached for {device.name}"}
            elif task_completed:
                self.__log.debug("RPC with method %s execution result is: %s", rpc_request.rpc_method, result)

            self.__gateway.send_rpc_reply(rpc_request.device_name,
                                          rpc_request.id,
                                          {"result": result})
            return result

        except Exception as e:
            self.__log.exception(e)
            self.__gateway.send_rpc_reply(rpc_request.device_name, rpc_request.id,
                                          {"result": {"error": str(e)}})

    def __process_reserved_rpc_request(self, rpc_request: OpcUaRpcRequest):
        identifier = ''
        device = self.__get_device_by_name(rpc_request.device_name)
        if device is None:
            self.__log.error('Device %s not found for RPC request', rpc_request.device_name)
            self.__gateway.send_rpc_reply(device=rpc_request.device_name,
                                          req_id=rpc_request.id,
                                          content={"result": {"error": 'Device not found'}})
            return
        is_node_id = self.__is_node_identifier(rpc_request.params)

        if is_node_id:
            identifier = rpc_request.params
            if not identifier:
                self.__log.error("Could not find node for requested rpc request %s", rpc_request.params)
                self.__gateway.send_rpc_reply(device=rpc_request.device_name,
                                              req_id=rpc_request.id,
                                              content={
                                                  "result": {"error": 'Could not find node for requested rpc request'}})

        elif not is_node_id:

            identifier = device.get_node_by_key(rpc_request.params)
            if not identifier:
                identifier = self.find_full_node_path(params=rpc_request.params, device=device)
            rpc_request.received_identifier = identifier

        try:
            task = self.__loop.create_task(self.__process_rpc_request(identifier=identifier, rpc_request=rpc_request))
            task_completed, result = self.__wait_task_with_timeout(task=task, timeout=rpc_request.timeout,
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.error(
                    "Failed to process rpc request for %s, timeout has been reached",
                    device.name,
                )
                result = {"error": f"Timeout rpc has been reached for {device.name}"}
            elif task_completed:
                self.__log.debug("RPC with method %s execution result is: %s", rpc_request.rpc_method, result)
            self.__gateway.send_rpc_reply(rpc_request.device_name,
                                          rpc_request.id,
                                          {"result": result})
            return result

        except Exception as e:
            self.__log.exception(e)
            self.__gateway.send_rpc_reply(rpc_request.device_name, rpc_request.id,
                                          {"result": {"error": str(e)}})

    async def __process_rpc_request(self, identifier: Node | str, rpc_request: OpcUaRpcRequest):
        result = {}
        try:
            if rpc_request.rpc_method == 'get':
                result = await self.__read_value(identifier)
                return result
            elif rpc_request.rpc_method == 'set':
                result = await self.__write_value(identifier, rpc_request.arguments)
                return result
            else:
                result['response'] = 'Unsupported function code in RPC request.'
                return result
        except Exception as e:
            self.__log.error('Failed to process rpc request: %r', e)
            result['response'] = {'error': e.__repr__()}
            return result

    async def __write_value(self, path, value):
        result = {}

        try:
            var = path
            if isinstance(path, str):
                var = self.__client.get_node(path.replace('\\.', '.'))
            elif isinstance(path, NodeId):
                var = self.__client.get_node(path)

            try:
                await var.write_value(value)
            except (BadWriteNotSupported, BadTypeMismatch):
                value = self.__guess_type_and_cast(value)
                data_value = ua.DataValue(ua.Variant(value))
                await var.write_value(data_value)

            result['value'] = value
            return result
        except UaStringParsingError:
            error_response = f"Could not find identifier in string {path}"
            result['error'] = error_response
            self.__log.error(error_response)
            return result
        except Exception as e:
            result['error'] = e.__str__()
            self.__log.error("Can not find node for provided path %s ", path)
            return result

    @staticmethod
    def __guess_type_and_cast(value):
        if isinstance(value, str):
            if value.lower() in ['true', 'false']:
                return value.lower() == 'true'

            try:
                if '.' in value:
                    return float(value)

                return int(value)
            except ValueError:
                return value

        return value

    async def __read_value(self, path):
        result = {}

        try:
            var = self.__client.get_node(path)
            result['value'] = await var.read_value()
            return result
        except UaStringParsingError:
            error_response = f"Could not find identifier in string {path}"
            result['error'] = error_response
            self.__log.error(error_response)
            return result
        except Exception as e:
            result['error'] = e.__str__()
            self.__log.error("Can not find node for provided path %s ", path)
            return result

    async def __call_method(self, path, method_name, arguments):

        result = {}
        try:
            var = await self.__client.nodes.root.get_child(path)
            method_id = '{}:{}'.format(var.nodeid.NamespaceIndex, method_name)
            result['result'] = await var.call_method(method_id, *arguments)
            return result
        except Exception as e:
            result['error'] = e.__str__()
            self.__log.error("Failed to execute rpc for %s Error: %s", method_name, e)
            self.__log.debug("Error", exc_info=e)
            return result

    async def __fetch_server_limitations(self):
        """Fetch and apply limitations from OPC-UA server capabilities."""
        try:
            server_caps = await self.__client.nodes.server.get_child("0:ServerCapabilities")
            if server_caps is None:
                self.__log.warning("ServerCapabilities node not found.")
                return
            nodes = await server_caps.get_children()

            for node in nodes:
                browse_name = await node.read_browse_name()
                name = browse_name.Name
                if name in {"MaxSessions", "MaxSubscriptions", "MaxBrowseContinuationPoints"}:
                    try:
                        val = await node.read_value()
                        self.__server_limits[name] = val
                    except Exception as e:
                        self.__log.warning("Failed to read %s: %s", name, e)

                elif name == "OperationLimits":
                    try:
                        op_limit_nodes = await node.get_children()
                        for op_node in op_limit_nodes:
                            op_name = (await op_node.read_browse_name()).Name
                            try:
                                val = await op_node.read_value()
                                self.__server_limits[f"OperationLimits.{op_name}"] = val
                            except Exception as e:
                                self.__log.warning("Failed to read OperationLimit %s: %s", op_name, e)
                    except Exception as e:
                        self.__log.warning("Failed to read OperationLimits node children: %s", e)

            self.__max_nodes_per_read = self.__server_limits.get("OperationLimits.MaxNodesPerRead", 100)
            self.__max_nodes_per_subscribe = self.__server_limits.get("OperationLimits.MaxNodesPerSubscribe", 100)

            self.__log.info("Applied server limitations: Read=%s, Subscribe=%s",
                            self.__max_nodes_per_read,
                            self.__max_nodes_per_subscribe)

        except Exception as e:
            self.__log.exception("Failed to fetch server limitations: %s", e)

    @staticmethod
    def __is_node_identifier(path):
        return isinstance(path, str) and re.match(r"(ns=\d+;[isgb]=[^}]+)", path)

    @staticmethod
    def __is_absolute_path(path):
        try:
            return path.replace('${', '').split('\\.')[0] == 'Root'
        except Exception:
            return False


class SubHandler:
    def __init__(self, queue, logger, status_change_callback):
        self.__log = logger
        self.__queue = queue
        self.__status_change_callback = status_change_callback

    def status_change_notification(self, status):
        self.__status_change_callback(status)

    def event_notification(self, event):
        self.__log.debug("New event %s", event)

    def datachange_notification(self, node, _, data):
        self.__log.trace("New data change event %s %s", node, data)
        self.__queue.put((node, data, int(time() * 1000)))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor, descriptor_pool as _descriptor_pool, \
    symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0emessages.proto\x12\x08messages\"\xa4\x01\n\x08Response\x12(\n\x06status\x18\x01 \x01(\x0e\x32\x18.messages.ResponseStatus\x12\x34\n\x0eserviceMessage\x18\x02 \x01(\x0b\x32\x1c.messages.FromServiceMessage\x12\x38\n\x10\x63onnectorMessage\x18\x03 \x01(\x0b\x32\x1e.messages.FromConnectorMessage\"\xa3\x05\n\x14\x46romConnectorMessage\x12$\n\x08response\x18\x01 \x01(\x0b\x32\x12.messages.Response\x12:\n\x13gatewayTelemetryMsg\x18\x02 \x01(\x0b\x32\x1d.messages.GatewayTelemetryMsg\x12<\n\x14gatewayAttributesMsg\x18\x03 \x01(\x0b\x32\x1e.messages.GatewayAttributesMsg\x12\x32\n\x0fgatewayClaimMsg\x18\x04 \x01(\x0b\x32\x19.messages.GatewayClaimMsg\x12<\n\x14registerConnectorMsg\x18\x05 \x01(\x0b\x32\x1e.messages.RegisterConnectorMsg\x12@\n\x16unregisterConnectorMsg\x18\x06 \x01(\x0b\x32 .messages.UnregisterConnectorMsg\x12(\n\nconnectMsg\x18\x07 \x01(\x0b\x32\x14.messages.ConnectMsg\x12.\n\rdisconnectMsg\x18\x08 \x01(\x0b\x32\x17.messages.DisconnectMsg\x12>\n\x15gatewayRpcResponseMsg\x18\t \x01(\x0b\x32\x1f.messages.GatewayRpcResponseMsg\x12I\n\x1agatewayAttributeRequestMsg\x18\n \x01(\x0b\x32%.messages.GatewayAttributesRequestMsg\x12R\n\x1f\x63onnectorGetConnectedDevicesMsg\x18\x0b \x01(\x0b\x32).messages.ConnectorGetConnectedDevicesMsg\"\x9e\x04\n\x12\x46romServiceMessage\x12$\n\x08response\x18\x01 \x01(\x0b\x32\x12.messages.Response\x12\x46\n\x19\x63onnectorConfigurationMsg\x18\x02 \x01(\x0b\x32#.messages.ConnectorConfigurationMsg\x12^\n%gatewayAttributeUpdateNotificationMsg\x18\x03 \x01(\x0b\x32/.messages.GatewayAttributeUpdateNotificationMsg\x12J\n\x1bgatewayAttributeResponseMsg\x18\x04 \x01(\x0b\x32%.messages.GatewayAttributeResponseMsg\x12H\n\x1agatewayDeviceRpcRequestMsg\x18\x05 \x01(\x0b\x32$.messages.GatewayDeviceRpcRequestMsg\x12@\n\x16unregisterConnectorMsg\x18\x06 \x01(\x0b\x32 .messages.UnregisterConnectorMsg\x12\x62\n\'connectorGetConnectedDevicesResponseMsg\x18\x07 \x01(\x0b\x32\x31.messages.ConnectorGetConnectedDevicesResponseMsg\",\n\x14RegisterConnectorMsg\x12\x14\n\x0c\x63onnectorKey\x18\x01 \x01(\t\".\n\x16UnregisterConnectorMsg\x12\x14\n\x0c\x63onnectorKey\x18\x01 \x01(\t\"^\n\x19\x43onnectorConfigurationMsg\x12\x15\n\rconnectorName\x18\x01 \x01(\t\x12\x15\n\rconfiguration\x18\x02 \x01(\t\x12\x13\n\x0b\x63onnectorId\x18\x03 \x01(\t\"7\n\x1f\x43onnectorGetConnectedDevicesMsg\x12\x14\n\x0c\x63onnectorKey\x18\x01 \x01(\t\"b\n\'ConnectorGetConnectedDevicesResponseMsg\x12\x37\n\x10\x63onnectorDevices\x18\x01 \x03(\x0b\x32\x1d.messages.ConnectorDeviceInfo\"=\n\x13\x43onnectorDeviceInfo\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\x12\n\ndeviceType\x18\x02 \x01(\t\"\x96\x01\n\rKeyValueProto\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x04type\x18\x02 \x01(\x0e\x32\x16.messages.KeyValueType\x12\x0e\n\x06\x62ool_v\x18\x03 \x01(\x08\x12\x0e\n\x06long_v\x18\x04 \x01(\x03\x12\x10\n\x08\x64ouble_v\x18\x05 \x01(\x01\x12\x10\n\x08string_v\x18\x06 \x01(\t\x12\x0e\n\x06json_v\x18\x07 \x01(\t\"<\n\tTsKvProto\x12\n\n\x02ts\x18\x01 \x01(\x03\x12#\n\x02kv\x18\x02 \x01(\x0b\x32\x17.messages.KeyValueProto\"@\n\rTsKvListProto\x12\n\n\x02ts\x18\x01 \x01(\x03\x12#\n\x02kv\x18\x02 \x03(\x0b\x32\x17.messages.KeyValueProto\"=\n\x10PostTelemetryMsg\x12)\n\x08tsKvList\x18\x01 \x03(\x0b\x32\x17.messages.TsKvListProto\"7\n\x10PostAttributeMsg\x12#\n\x02kv\x18\x01 \x03(\x0b\x32\x17.messages.KeyValueProto\"c\n\x1e\x41ttributeUpdateNotificationMsg\x12*\n\rsharedUpdated\x18\x01 \x03(\x0b\x32\x13.messages.TsKvProto\x12\x15\n\rsharedDeleted\x18\x02 \x03(\t\"N\n\x15ToDeviceRpcRequestMsg\x12\x11\n\trequestId\x18\x01 \x01(\x05\x12\x12\n\nmethodName\x18\x02 \x01(\t\x12\x0e\n\x06params\x18\x03 \x01(\t\"K\n\x16ToDeviceRpcResponseMsg\x12\x11\n\trequestId\x18\x01 \x01(\x05\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"N\n\x15ToServerRpcRequestMsg\x12\x11\n\trequestId\x18\x01 \x01(\x05\x12\x12\n\nmethodName\x18\x02 \x01(\t\x12\x0e\n\x06params\x18\x03 \x01(\t\"K\n\x16ToServerRpcResponseMsg\x12\x11\n\trequestId\x18\x01 \x01(\x05\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"4\n\x0b\x43laimDevice\x12\x11\n\tsecretKey\x18\x01 \x01(\t\x12\x12\n\ndurationMs\x18\x02 \x01(\x03\";\n\x11\x41ttributesRequest\x12\x12\n\nclientKeys\x18\x01 \x01(\t\x12\x12\n\nsharedKeys\x18\x02 \x01(\t\",\n\nRpcRequest\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x01(\t\"#\n\rDisconnectMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\"4\n\nConnectMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\x12\n\ndeviceType\x18\x02 \x01(\t\"K\n\x0cTelemetryMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\'\n\x03msg\x18\x03 \x01(\x0b\x32\x1a.messages.PostTelemetryMsg\"L\n\rAttributesMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\'\n\x03msg\x18\x02 \x01(\x0b\x32\x1a.messages.PostAttributeMsg\"Q\n\x0e\x43laimDeviceMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12+\n\x0c\x63laimRequest\x18\x02 \x01(\x0b\x32\x15.messages.ClaimDevice\":\n\x13GatewayTelemetryMsg\x12#\n\x03msg\x18\x01 \x03(\x0b\x32\x16.messages.TelemetryMsg\"8\n\x0fGatewayClaimMsg\x12%\n\x03msg\x18\x01 \x03(\x0b\x32\x18.messages.ClaimDeviceMsg\"<\n\x14GatewayAttributesMsg\x12$\n\x03msg\x18\x01 \x03(\x0b\x32\x17.messages.AttributesMsg\"E\n\x15GatewayRpcResponseMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t\"n\n\x1bGatewayAttributeResponseMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12;\n\x0bresponseMsg\x18\x02 \x01(\x0b\x32&.messages.GatewayAttributesResponseMsg\"~\n%GatewayAttributeUpdateNotificationMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\x41\n\x0fnotificationMsg\x18\x02 \x01(\x0b\x32(.messages.AttributeUpdateNotificationMsg\"h\n\x1aGatewayDeviceRpcRequestMsg\x12\x12\n\ndeviceName\x18\x01 \x01(\t\x12\x36\n\rrpcRequestMsg\x18\x02 \x01(\x0b\x32\x1f.messages.ToDeviceRpcRequestMsg\"[\n\x1bGatewayAttributesRequestMsg\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x0e\n\x06\x63lient\x18\x03 \x01(\x08\x12\x0c\n\x04keys\x18\x04 \x03(\t\"\xac\x01\n\x1cGatewayAttributesResponseMsg\x12\x11\n\trequestId\x18\x01 \x01(\x05\x12\x34\n\x13\x63lientAttributeList\x18\x02 \x03(\x0b\x32\x17.messages.KeyValueProto\x12\x34\n\x13sharedAttributeList\x18\x03 \x03(\x0b\x32\x17.messages.KeyValueProto\x12\r\n\x05\x65rror\x18\x05 \x01(\t*F\n\x0eResponseStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\r\n\tNOT_FOUND\x10\x02\x12\x0b\n\x07\x46\x41ILURE\x10\x03*Q\n\x0cKeyValueType\x12\r\n\tBOOLEAN_V\x10\x00\x12\n\n\x06LONG_V\x10\x01\x12\x0c\n\x08\x44OUBLE_V\x10\x02\x12\x0c\n\x08STRING_V\x10\x03\x12\n\n\x06JSON_V\x10\x04\x32_\n\x15TBGatewayProtoService\x12\x46\n\x06stream\x12\x1e.messages.FromConnectorMessage\x1a\x1c.messages.FromServiceMessageb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'messages_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals['_RESPONSESTATUS']._serialized_start = 3999
    _globals['_RESPONSESTATUS']._serialized_end = 4069
    _globals['_KEYVALUETYPE']._serialized_start = 4071
    _globals['_KEYVALUETYPE']._serialized_end = 4152
    _globals['_RESPONSE']._serialized_start = 29
    _globals['_RESPONSE']._serialized_end = 193
    _globals['_FROMCONNECTORMESSAGE']._serialized_start = 196
    _globals['_FROMCONNECTORMESSAGE']._serialized_end = 871
    _globals['_FROMSERVICEMESSAGE']._serialized_start = 874
    _globals['_FROMSERVICEMESSAGE']._serialized_end = 1416
    _globals['_REGISTERCONNECTORMSG']._serialized_start = 1418
    _globals['_REGISTERCONNECTORMSG']._serialized_end = 1462
    _globals['_UNREGISTERCONNECTORMSG']._serialized_start = 1464
    _globals['_UNREGISTERCONNECTORMSG']._serialized_end = 1510
    _globals['_CONNECTORCONFIGURATIONMSG']._serialized_start = 1512
    _globals['_CONNECTORCONFIGURATIONMSG']._serialized_end = 1606
    _globals['_CONNECTORGETCONNECTEDDEVICESMSG']._serialized_start = 1608
    _globals['_CONNECTORGETCONNECTEDDEVICESMSG']._serialized_end = 1663
    _globals['_CONNECTORGETCONNECTEDDEVICESRESPONSEMSG']._serialized_start = 1665
    _globals['_CONNECTORGETCONNECTEDDEVICESRESPONSEMSG']._serialized_end = 1763
    _globals['_CONNECTORDEVICEINFO']._serialized_start = 1765
    _globals['_CONNECTORDEVICEINFO']._serialized_end = 1826
    _globals['_KEYVALUEPROTO']._serialized_start = 1829
    _globals['_KEYVALUEPROTO']._serialized_end = 1979
    _globals['_TSKVPROTO']._serialized_start = 1981
    _globals['_TSKVPROTO']._serialized_end = 2041
    _globals['_TSKVLISTPROTO']._serialized_start = 2043
    _globals['_TSKVLISTPROTO']._serialized_end = 2107
    _globals['_POSTTELEMETRYMSG']._serialized_start = 2109
    _globals['_POSTTELEMETRYMSG']._serialized_end = 2170
    _globals['_POSTATTRIBUTEMSG']._serialized_start = 2172
    _globals['_POSTATTRIBUTEMSG']._serialized_end = 2227
    _globals['_ATTRIBUTEUPDATENOTIFICATIONMSG']._serialized_start = 2229
    _globals['_ATTRIBUTEUPDATENOTIFICATIONMSG']._serialized_end = 2328
    _globals['_TODEVICERPCREQUESTMSG']._serialized_start = 2330
    _globals['_TODEVICERPCREQUESTMSG']._serialized_end = 2408
    _globals['_TODEVICERPCRESPONSEMSG']._serialized_start = 2410
    _globals['_TODEVICERPCRESPONSEMSG']._serialized_end = 2485
    _globals['_TOSERVERRPCREQUESTMSG']._serialized_start = 2487
    _globals['_TOSERVERRPCREQUESTMSG']._serialized_end = 2565
    _globals['_TOSERVERRPCRESPONSEMSG']._serialized_start = 2567
    _globals['_TOSERVERRPCRESPONSEMSG']._serialized_end = 2642
    _globals['_CLAIMDEVICE']._serialized_start = 2644
    _globals['_CLAIMDEVICE']._serialized_end = 2696
    _globals['_ATTRIBUTESREQUEST']._serialized_start = 2698
    _globals['_ATTRIBUTESREQUEST']._serialized_end = 2757
    _globals['_RPCREQUEST']._serialized_start = 2759
    _globals['_RPCREQUEST']._serialized_end = 2803
    _globals['_DISCONNECTMSG']._serialized_start = 2805
    _globals['_DISCONNECTMSG']._serialized_end = 2840
    _globals['_CONNECTMSG']._serialized_start = 2842
    _globals['_CONNECTMSG']._serialized_end = 2894
    _globals['_TELEMETRYMSG']._serialized_start = 2896
    _globals['_TELEMETRYMSG']._serialized_end = 2971
    _globals['_ATTRIBUTESMSG']._serialized_start = 2973
    _globals['_ATTRIBUTESMSG']._serialized_end = 3049
    _globals['_CLAIMDEVICEMSG']._serialized_start = 3051
    _globals['_CLAIMDEVICEMSG']._serialized_end = 3132
    _globals['_GATEWAYTELEMETRYMSG']._serialized_start = 3134
    _globals['_GATEWAYTELEMETRYMSG']._serialized_end = 3192
    _globals['_GATEWAYCLAIMMSG']._serialized_start = 3194
    _globals['_GATEWAYCLAIMMSG']._serialized_end = 3250
    _globals['_GATEWAYATTRIBUTESMSG']._serialized_start = 3252
    _globals['_GATEWAYATTRIBUTESMSG']._serialized_end = 3312
    _globals['_GATEWAYRPCRESPONSEMSG']._serialized_start = 3314
    _globals['_GATEWAYRPCRESPONSEMSG']._serialized_end = 3383
    _globals['_GATEWAYATTRIBUTERESPONSEMSG']._serialized_start = 3385
    _globals['_GATEWAYATTRIBUTERESPONSEMSG']._serialized_end = 3495
    _globals['_GATEWAYATTRIBUTEUPDATENOTIFICATIONMSG']._serialized_start = 3497
    _globals['_GATEWAYATTRIBUTEUPDATENOTIFICATIONMSG']._serialized_end = 3623
    _globals['_GATEWAYDEVICERPCREQUESTMSG']._serialized_start = 3625
    _globals['_GATEWAYDEVICERPCREQUESTMSG']._serialized_end = 3729
    _globals['_GATEWAYATTRIBUTESREQUESTMSG']._serialized_start = 3731
    _globals['_GATEWAYATTRIBUTESREQUESTMSG']._serialized_end = 3822
    _globals['_GATEWAYATTRIBUTESRESPONSEMSG']._serialized_start = 3825
    _globals['_GATEWAYATTRIBUTESRESPONSEMSG']._serialized_end = 3997
    _globals['_TBGATEWAYPROTOSERVICE']._serialized_start = 4154
    _globals['_TBGATEWAYPROTOSERVICE']._serialized_end = 4249
# @@protoc_insertion_point(module_scope)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from simplejson import dumps, loads

from thingsboard_gateway.connectors.request.request_converter import RequestConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService


class JsonRequestUplinkConverter(RequestConverter):
    def __init__(self, config, logger):
        self.__log = logger
        self.__config = config
        self.__datatypes = {"attributes": "attributes",
                            "telemetry": "telemetry"}

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        device_name = None
        device_type = None

        if isinstance(data, (bytes, str)):
            data = loads(data)

        try:
            if self.__config['converter'].get("deviceNameJsonExpression") is not None:
                device_name_tags = TBUtility.get_values(self.__config['converter'].get("deviceNameJsonExpression"),
                                                        data, get_tag=True)
                device_name_values = TBUtility.get_values(self.__config['converter'].get("deviceNameJsonExpression"),
                                                          data, get_tag=False)
                device_name = self.__config['converter'].get("deviceNameJsonExpression")
                for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                    is_valid_key = "${" in self.__config['converter'].get("deviceNameJsonExpression") and "}" in \
                                   self.__config['converter'].get("deviceNameJsonExpression")
                    device_name = device_name.replace('${' + str(device_name_tag) + '}', str(device_name_value)) \
                        if is_valid_key else device_name
            else:
                self.__log.error("The expression for looking \"deviceName\" not found in config %s",
                                 dumps(self.__config['converter']))
            if self.__config['converter'].get("deviceTypeJsonExpression") is not None:
                device_type_tags = TBUtility.get_values(self.__config['converter'].get("deviceTypeJsonExpression"),
                                                        data,
                                                        get_tag=True)
                device_type_values = TBUtility.get_values(self.__config['converter'].get("deviceTypeJsonExpression"),
                                                          data,
                                                          expression_instead_none=True)
                device_type = self.__config['converter'].get("deviceTypeJsonExpression")
                for (device_type_tag, device_type_value) in zip(device_type_tags, device_type_values):
                    is_valid_key = "${" in self.__config['converter'].get("deviceTypeJsonExpression") and "}" in \
                                   self.__config['converter'].get("deviceTypeJsonExpression")
                    device_type = device_type.replace('${' + str(device_type_tag) + '}', str(device_type_value)) \
                        if is_valid_key else device_type_tag
            else:
                self.__log.error("The expression for looking \"deviceType\" not found in config %s",
                                 dumps(self.__config['converter']))
        except Exception as e:
            StatisticsService.count_connector_message(self.__log.name, 'convertersMsgDropped')
            self.__log.exception(e)

        converted_data = ConvertedData(device_name=device_name, device_type=device_type)
        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self.__log.trace("Report strategy config is not specified for device %s: %s", device_name, e)

        try:
            for datatype in self.__datatypes:
                for datatype_object_config in self.__config["converter"].get(datatype, []):
                    values = TBUtility.get_values(datatype_object_config["value"], data, datatype_object_config["type"],
                                                  expression_instead_none=True)
                    values_tags = TBUtility.get_values(datatype_object_config["value"], data,
                                                       datatype_object_config["type"],
                                                       get_tag=True)

                    keys = TBUtility.get_values(datatype_object_config["key"], data, datatype_object_config["type"],
                                                expression_instead_none=True)
                    keys_tags = TBUtility.get_values(datatype_object_config["key"], data, get_tag=True)

                    full_key = datatype_object_config["key"]
                    for (key, key_tag) in zip(keys, keys_tags):
                        is_valid_key = "${" in datatype_object_config["key"] and "}" in \
                                       datatype_object_config["key"]
                        full_key = full_key.replace('${' + str(key_tag) + '}',
                                                    str(key)) if is_valid_key else key_tag

                    full_value = datatype_object_config["value"]
                    for (value, value_tag) in zip(values, values_tags):
                        is_valid_value = "${" in datatype_object_config["value"] and "}" in \
                                         datatype_object_config["value"]

                        full_value = full_value.replace('${' + str(value_tag) + '}',
                                                        str(value)) if is_valid_value else str(value)

                    datapoint_key = TBUtility.convert_key_to_datapoint_key(full_key, device_report_strategy,
                                                                           datatype_object_config, self.__log)
                    if datatype == 'attributes':
                        converted_data.add_to_attributes(datapoint_key, full_value)
                    else:
                        ts = TBUtility.resolve_different_ts_formats(data=data, config=datatype_object_config, logger=self.__log)

                        telemetry_entry = TelemetryEntry({datapoint_key: full_value}, ts)
                        converted_data.add_to_telemetry(telemetry_entry)
        except Exception as e:
            StatisticsService.count_connector_message(self.__log.name, 'convertersMsgDropped')
            self.__log.exception(e)

        StatisticsService.count_connector_message(self.__log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self.__log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        self.__log.debug("Converted data: %s", converted_data)

        return converted_data



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from logging import getLogger
from queue import Empty, Full, Queue

from thingsboard_gateway.storage.event_storage import EventStorage


class MemoryEventStorage(EventStorage):
    def __init__(self, config, logger, main_stop_event):
        super().__init__(config, logger, main_stop_event)
        self.__log = logger
        self.__queue_len = config.get("max_records_count", 10000)
        self.__events_per_time = config.get("read_records_count", 1000)
        self.__events_queue = Queue(self.__queue_len)
        self.__event_pack = []
        self.__stopped = False
        self.__log.debug("Memory storage created with following configuration: \nMax size: %i\n Read records per time: %i",
                  self.__queue_len, self.__events_per_time)

    def put(self, event):
        success = False
        if not self.__stopped:
            try:
                self.__events_queue.put_nowait(event)
                success = True
            except Full:
                self.__log.error("Memory storage is full!")
        else:
            self.__log.error("Storage is stopped!")
        return success

    def get_event_pack(self):
        try:
            if not self.__event_pack:
                self.__event_pack = [self.__events_queue.get_nowait() for _ in
                                     range(min(self.__events_per_time, self.__events_queue.qsize()))]
        except Empty:
            pass
        return self.__event_pack

    def event_pack_processing_done(self):
        self.__event_pack = []

    def stop(self):
        self.__stopped = True

    def len(self):
        return self.__events_queue.qsize()

    def update_logger(self):
        self.__log = getLogger("storage")



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from ast import literal_eval
from urllib.parse import quote

from simplejson import dumps

from thingsboard_gateway.connectors.request.request_converter import RequestConverter
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class JsonRequestDownlinkConverter(RequestConverter):
    def __init__(self, config, logger):
        self.__log = logger
        self.__config = config

    @CollectStatistics(start_stat_type='allReceivedBytesFromTB',
                       end_stat_type='allBytesSentToDevices')
    def convert(self, config, data):
        try:
            if data["data"].get("id") is None:
                attribute_key = list(data["data"].keys())[0]
                attribute_value = list(data["data"].values())[0]

                result = {
                    "url": self.__config["requestUrlExpression"].replace("${attributeKey}", quote(attribute_key))
                    .replace("${attributeValue}", quote(str(attribute_value)))
                    .replace("${deviceName}", quote(data["device"])),
                    "data": self.__config["requestValueExpression"].replace("${attributeKey}", attribute_key)
                    .replace("${attributeValue}", str(attribute_value))
                    .replace("${deviceName}", data["device"])
                }
            else:
                request_id = str(data["data"]["id"])
                method_name = data["data"]["method"]

                result = {
                    "url": self.__config["requestUrlExpression"].replace("${requestId}", request_id)
                    .replace("${methodName}", method_name)
                    .replace("${deviceName}", quote(data["device"])),
                    "data": self.__config["requestValueExpression"].replace("${requestId}", request_id)
                    .replace("${methodName}", method_name)
                    .replace("${deviceName}", data["device"])
                }

            result['url'] = TBUtility.replace_params_tags(result['url'], data)

            data_tags = TBUtility.get_values(config.get('requestValueExpression'), data['data'], 'params',
                                             get_tag=True)
            data_values = TBUtility.get_values(config.get('requestValueExpression'), data['data'], 'params',
                                               expression_instead_none=True)

            for (tag, value) in zip(data_tags, data_values):
                result['data'] = result["data"].replace('${' + tag + '}', str(value))

            result["data"] = dumps(literal_eval(result["data"]))
            return result
        except Exception as e:
            self.__log.exception(e)



from copy import deepcopy
import re
from typing import List

from thingsboard_gateway.tb_utility.tb_logger import TbLogger


class BackwardCompatibilityAdapter:

    def __init__(self, config, logger: TbLogger):
        self._config = deepcopy(config)
        self._log = logger

    def convert(self):
        # Move server-related keys under 'server'
        server_keys = {'host', 'port', 'SSL', 'security'}
        self._config['server'] = {k: self._config.pop(k) for k in list(self._config) if k in server_keys}

        # Process and convert 'mapping' section
        old_mappings = self._config.get('mapping')

        new_mappings = []
        if isinstance(old_mappings, list):
            for mapping in deepcopy(old_mappings):
                converter = mapping.get('converter', {})
                if not isinstance(converter, dict):
                    self._log.error("Invalid converter format in mapping: %r", mapping)
                    continue

                # Convert device info block
                device_name_expr = converter.pop('deviceNameExpression', None)
                device_type_expr = converter.pop('deviceTypeExpression', 'default')

                converter['deviceInfo'] = {
                    'deviceNameExpressionSource': self.get_value_source(device_name_expr),
                    'deviceNameExpression': device_name_expr,
                    'deviceProfileExpressionSource': self.get_value_source(device_type_expr),
                    'deviceProfileExpression': device_type_expr
                }

                # Rename extension-config → extensionConfig (for custom converters)
                if 'extension-config' in converter:
                    converter['extensionConfig'] = converter.pop('extension-config')

                mapping['converter'] = converter
                new_mappings.append(mapping)
        else:
            self._log.error("Invalid 'mapping' format: %r", old_mappings)

        self._config['mapping'] = new_mappings

        # Move requests-related sections under 'requestsMapping'
        requests_mapping = {}
        for key in ('attributeRequests', 'attributeUpdates', 'serverSideRpc'):
            value = self._config.pop(key, None)
            if value is not None:
                requests_mapping[key] = value

        if isinstance(self._config.get('requestsMapping'), list):
            self._config['requestsMapping'].extend(requests_mapping)
        else:
            self._config['requestsMapping'] = requests_mapping

        return self._config

    @staticmethod
    def get_value_source(value, possible_constant=True):
        if re.search(r"\${([A-Za-z.:\\\d]+)}", value) or not possible_constant:
            return 'request'
        else:
            return 'constant'

    @staticmethod
    def is_old_config(config):
        return 'server' not in config or 'requestsMapping' not in config



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from re import search
from time import time
from typing import Dict

from simplejson import dumps

from thingsboard_gateway.connectors.mqtt.mqtt_uplink_converter import MqttUplinkConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER, \
    RECEIVED_TS_PARAMETER, CONVERTED_TS_PARAMETER
from thingsboard_gateway.gateway.entities.attributes import Attributes
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService

USE_RECEIVED_TS_PARAMETER = "useReceivedTs"


class JsonMqttUplinkConverter(MqttUplinkConverter):
    CONFIGURATION_OPTION_USE_EVAL = "useEval"

    def __init__(self, config, logger):
        self._log = logger
        self.__device_report_strategy = None
        try:
            self.__device_report_strategy = ReportStrategyConfig(config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            pass
        self.__config = config.get('converter')
        self.__use_eval = self.__config.get(self.CONFIGURATION_OPTION_USE_EVAL, False)

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        self.__config = value

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, topic, data):
        StatisticsService.count_connector_message(self._log.name, 'convertersMsgProcessed')

        if isinstance(data, list):
            converted_data_devices: Dict[str, ConvertedData] = {}
            for item in data:
                converted_item = self._convert_single_item(topic, item)
                if converted_item.device_name not in converted_data_devices:
                    converted_data_devices[converted_item.device_name] = converted_item
                else:
                    existing_item = converted_data_devices[converted_item.device_name]
                    existing_item.extend(converted_item)
            self._log.debug("Converted data for %s devices", len(converted_data_devices))
            return list(converted_data_devices.values())
        else:
            return self._convert_single_item(topic, data)

    def _convert_single_item(self, topic, data):
        datatypes = {"attributes": "attributes",
                     "timeseries": "telemetry"}

        device_name = self.parse_device_name(topic, data, self.__config)

        converted_data = ConvertedData(device_name=device_name,
                                       device_type=self.parse_device_type(topic, data, self.__config),
                                       metadata={RECEIVED_TS_PARAMETER: int(time() * 1000)})

        try:
            for datatype in datatypes:
                timestamp = None
                if self.__config.get(USE_RECEIVED_TS_PARAMETER, False) is True:
                    timestamp = converted_data.metadata["receivedTs"]

                for datatype_config in self.__config.get(datatype, []):
                    if isinstance(datatype_config, str) and datatype_config == "*":
                        if datatype == "attributes":
                            converted_data.add_to_attributes(Attributes(data))
                        else:
                            telemetry_entry = TelemetryEntry(data, timestamp)
                            converted_data.add_to_telemetry(telemetry_entry)
                    else:
                        values = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                      expression_instead_none=False)
                        values_tags = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                           get_tag=True)

                        keys = TBUtility.get_values(datatype_config["key"], data, datatype_config["type"],
                                                    expression_instead_none=False)
                        keys_tags = TBUtility.get_values(datatype_config["key"], data, get_tag=True)

                        full_key = datatype_config["key"]
                        for (key, key_tag) in zip(keys, keys_tags):
                            is_valid_key = "${" in datatype_config["key"] and "}" in datatype_config["key"]
                            full_key = full_key.replace('${' + str(key_tag) + '}', str(key)) if is_valid_key else key_tag

                        full_value = datatype_config["value"]
                        for (value, value_tag) in zip(values, values_tags):
                            is_valid_value = "${" in datatype_config["value"] and "}" in datatype_config["value"]
                            full_value = full_value.replace('${' + str(value_tag) + '}', str(value)) if is_valid_value else value

                        if full_key != 'None' and full_value != 'None':
                            converted_key = TBUtility.convert_key_to_datapoint_key(full_key, self.__device_report_strategy, datatype_config, self._log)
                            converted_value = TBUtility.convert_data_type(full_value, datatype_config["type"], self.__use_eval)
                            if datatype == "attributes":
                                converted_data.add_to_attributes(converted_key, converted_value)
                            else:
                                if timestamp is None:
                                    timestamp = TBUtility.resolve_different_ts_formats(data=data, config=datatype_config, logger=self._log)
                                telemetry_entry = TelemetryEntry({converted_key: converted_value}, timestamp)
                                converted_data.add_to_telemetry(telemetry_entry)
        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            str(data), e)
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')

        self._log.debug("Converted data: %s", converted_data)

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)
        return converted_data

    @staticmethod
    def create_data_record(key, value, timestamp):
        value_item = {key: value}
        return {"ts": timestamp, 'values': value_item} if timestamp else value_item

    def parse_device_name(self, topic, data, config):
        return self.parse_device_info(
            topic, data, config, "deviceNameExpressionSource", "deviceNameExpression")

    def parse_device_type(self, topic, data, config):
        return self.parse_device_info(
            topic, data, config, "deviceProfileExpressionSource", "deviceProfileExpression")

    def parse_device_info(self, topic, data, config, expression_source, expression):
        result = None
        device_info = config.get('deviceInfo', {})

        expression = device_info.get('deviceNameExpression') if expression == 'deviceNameExpression' \
            else device_info.get('deviceProfileExpression')

        try:
            if device_info.get(expression_source) == 'message' or device_info.get(expression_source) == 'constant':
                result_tags = TBUtility.get_values(expression, data, get_tag=True)
                result_values = TBUtility.get_values(expression, data, expression_instead_none=True)

                result = expression
                for (result_tag, result_value) in zip(result_tags, result_values):
                    is_valid_key = "${" in expression and "}" in expression
                    result = result.replace('${' + str(result_tag) + '}',
                                            str(result_value)) if is_valid_key else result_tag
            elif device_info.get(expression_source) == 'topic':
                search_result = search(expression, topic)
                if search_result is not None:
                    result = search_result.group(0)
                else:
                    self._log.debug(
                        "Regular expression result is None. deviceNameTopicExpression parameter will be interpreted "
                        "as a deviceName\n Topic: %s\nRegex: %s", topic, expression)
                    result = expression
            else:
                self._log.error("The expression for looking \"deviceName\" not found in config %s", dumps(config))
        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(config), data, e)
        return result



# ------------------------------------------------------------------------------
#      Copyright 2025. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
#
# ------------------------------------------------------------------------------

from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig


class DatapointKey:
    def __init__(self, key, report_strategy: ReportStrategyConfig = None):
        self.key = key
        self.report_strategy = report_strategy

    def __str__(self):
        return f"DatapointKey(key={self.key}, report_strategy={self.report_strategy})"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self.key, self.report_strategy))

    def __eq__(self, other):
        if isinstance(other, DatapointKey):
            return self.key == other.key and self.report_strategy == other.report_strategy
        return False



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from copy import deepcopy


class FTPBackwardCompatibilityAdapter:

    def __init__(self, config: dict) -> None:
        self._config = deepcopy(config)
        self._converted_config = {
            "parameters": {
            },
            "paths": [],
            "requestsMapping": {
                "attributeUpdates": [],
                "serverSideRpc": []
            }
        }

    def convert(self) -> dict:
        self._converted_config['id'] = self._config.get('id')
        self._converted_config['name'] = self._config.get('name')
        self._converted_config['logLevel'] = self._config.get('logLevel', 'INFO')
        self._converted_config['enableRemoteLogging'] = self._config.get('enableRemoteLogging', False)
        self._converted_config['configVersion'] = self._config.get('configVersion')

        self._converted_config['parameters']['host'] = self._config.get('host')
        self._converted_config['parameters']['port'] = self._config.get('port')
        self._converted_config["parameters"]["TLSSupport"] = self._config.get("TLSSupport", False)
        security = self._config.get('security', {})
        if not security.get('type') == 'basic':
            security.pop("username", None)
            security.pop("password", None)
        self._converted_config['parameters']['security'] = security
        old_paths = self._config.get('paths', [])
        new_paths = self._convert_path_entries(old_paths)
        self._converted_config["paths"] = new_paths
        self._converted_config["requestsMapping"]["attributeUpdates"] = self._config.get("attributeUpdates", [])
        self._converted_config["requestsMapping"]["serverSideRpc"] = self._config.get("serverSideRpc", [])

        return self._converted_config

    @staticmethod
    def _convert_path_entries(old_paths: list) -> list:
        new_paths = []
        for old_path_configuration in old_paths:
            new_path_configuration = FTPBackwardCompatibilityAdapter._convert_path(old_path_configuration)
            new_paths.append(new_path_configuration)
        return new_paths

    @staticmethod
    def _convert_path(old_path_config: dict) -> dict:
        converted_configuration = {}
        for key, value in old_path_config.items():
            converted_configuration[key] = old_path_config[key]

        if "attributes" in old_path_config:
            converted_configuration["attributes"] = []
            for attribute in old_path_config["attributes"]:
                new_attribute = {}
                attribute_type = attribute.get("type", "string")
                if attribute_type == "int":
                    attribute_type = "integer"
                elif attribute_type == "string":
                    attribute_type = "string"

                new_attribute["type"] = attribute_type

                for key, value in attribute.items():
                    if key != "type":
                        new_attribute[key] = value

                converted_configuration["attributes"].append(new_attribute)

            if "timeseries" in old_path_config:
                converted_configuration["timeseries"] = []
                for ts in old_path_config["timeseries"]:
                    new_ts = ts
                    if new_ts.get("type") == "int":
                        new_ts["type"] = "integer"
                    elif new_ts.get("type") == "str":
                        new_ts["type"] = "string"

                    converted_configuration["timeseries"].append(new_ts)

        return converted_configuration

    @staticmethod
    def is_old_config_format(config: dict) -> bool:
        return True if not config.get("parameters") else False



from copy import deepcopy


class BackwardCompatibilityAdapter:
    def __init__(self, config):
        self._config = deepcopy(config)

    def convert(self):
        self._config['requestsMapping'] = {}
        config_mapping = {'requestsMapping': ('connectRequests', 'disconnectRequests', 'attributeRequests',
                                       'attributeUpdates', 'serverSideRpc'),
                   'mapping': ('mapping', 'dataMapping')}

        has_mapping_section = False
        for (map_section, map_type) in config_mapping.items():
            for t in map_type:
                section_config = self._config.pop(t, {})
                try:
                    for item in section_config:
                        (device_name_json_expression, device_type_json_expression, device_name_topic_expression,
                         device_type_topic_expression, bytes_converter) = self._get_device_name_and_type(item)

                        self._parse_device_info(item['converter'] if item.get('converter') else item, device_name_json_expression,
                                                device_type_json_expression,
                                                device_name_topic_expression, device_type_topic_expression,
                                                bytes_converter)

                        if t == 'attributeRequests':
                            self._parce_attribute_info(item)
                    if not has_mapping_section and (t == 'mapping' or t == 'dataMapping') and section_config:
                        self._config[map_section] = section_config
                        has_mapping_section = True
                        break
                    else:
                        self._config[map_section][t] = section_config
                except KeyError:
                    continue
                except AttributeError as e:
                    print('Can\'t parce config section "{}", subsection "{}". '
                          'Please, check your configuration.\nError: {}'
                          .format(t, section_config, e))
                    continue

        if not self._config.get('mapping'):
            self._config['mapping'] = []

        return self._config

    @staticmethod
    def _get_device_name_and_type(config):
        if config.get('converter', {}).get('deviceNameExpression') and config.get('converter', {}).get(
                'deviceTypeExpression'):
            device_name_json_expression = config.get('converter', {}).pop('deviceNameExpression', None)
            device_type_json_expression = config.get('converter', {}).pop('deviceTypeExpression', None)

            return device_name_json_expression, device_type_json_expression, None, None, True

        device_name_json_expression = config.get('converter', config).pop('deviceNameJsonExpression', None)
        device_type_json_expression = config.get('converter', config).pop('deviceTypeJsonExpression', None)

        device_name_topic_expression = config.get('converter', config).pop('deviceNameTopicExpression', None)
        device_type_topic_expression = config.get('converter', config).pop('deviceTypeTopicExpression', None)

        return (device_name_json_expression, device_type_json_expression,
                device_name_topic_expression, device_type_topic_expression, False)

    @staticmethod
    def _parse_device_info(config, device_name_json_expression=None, device_type_json_expression=None,
                           device_name_topic_expression=None, device_type_topic_expression=None, bytes_converter=False):
        if bytes_converter:
            config['deviceInfo'] = {}
            config['deviceInfo']['deviceNameExpression'] = device_name_json_expression
            config['deviceInfo']['deviceProfileExpression'] = device_type_json_expression
            return
        if config.get('deviceInfo') is None:
            config['deviceInfo'] = {}

        if device_name_json_expression:
            config['deviceInfo']['deviceNameExpressionSource'] = 'message'
            config['deviceInfo']['deviceNameExpression'] = device_name_json_expression

        if device_type_json_expression:
            config['deviceInfo']['deviceProfileExpressionSource'] = 'message'
            config['deviceInfo']['deviceProfileExpression'] = device_type_json_expression

        if device_name_topic_expression:
            config['deviceInfo']['deviceNameExpressionSource'] = 'topic'
            config['deviceInfo']['deviceNameExpression'] = device_name_topic_expression

        if device_type_topic_expression:
            config['deviceInfo']['deviceProfileExpressionSource'] = 'topic'
            config['deviceInfo']['deviceProfileExpression'] = device_type_topic_expression

        if config.get('extension-config'):
            extension_config = config.pop('extension-config')
            config['extensionConfig'] = extension_config

        if len(config['deviceInfo']) == 0:
            config.pop('deviceInfo')

    @staticmethod
    def _parce_attribute_info(config):
        attribute_name_expression = config.pop('attributeNameJsonExpression', None)
        if attribute_name_expression:
            config['attributeNameExpressionSource'] = 'message'
            config['attributeNameExpression'] = attribute_name_expression
            return

        attribute_name_expression = config.pop('attributeNameTopicExpression', None)
        if attribute_name_expression:
            config['attributeNameExpressionSource'] = 'topic'
            config['attributeNameExpression'] = attribute_name_expression

    @staticmethod
    def is_old_config_format(config):
        mapping = config.get('mapping')
        return mapping is not None and len(mapping) > 0 and mapping[0].get('converter', {}).get('deviceInfo') is None



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from time import monotonic
from asyncio import Lock

from pymodbus.client import AsyncModbusSerialClient
from pymodbus.client.tcp import AsyncModbusTcpClient
from pymodbus.client.udp import AsyncModbusUdpClient
from pymodbus.framer.base import FramerType

from thingsboard_gateway.connectors.modbus.entities.clients import AsyncModbusTlsClient
from thingsboard_gateway.connectors.modbus.constants import SERIAL_CONNECTION_TYPE_PARAMETER


def with_lock_for_serial(func):
    async def wrapper(master, *args, **kwargs):
        if master.client_type == SERIAL_CONNECTION_TYPE_PARAMETER:
            await master.lock.acquire()
        try:
            resp = await func(master, *args, **kwargs)
        finally:
            if master.client_type == SERIAL_CONNECTION_TYPE_PARAMETER:
                master.lock.release()

        return resp

    return wrapper


class Master:
    def __init__(self, client_type, client):
        self.lock = Lock()
        self.client_type = client_type.lower()
        self.__client = client
        self.__previous_request_time = 0

    def get_time_to_pass_delay_between_requests(self, delay_ms) -> int:
        if delay_ms == 0:
            return 0
        next_possible_request_time = self.__previous_request_time + delay_ms
        current_time = int(monotonic() * 1000)
        if current_time >= next_possible_request_time:
            return 0
        return next_possible_request_time - current_time

    def connected(self):
        return self.__client.connected

    async def connect(self):
        """
        Connect to the Modbus client if client is not connected.

        Lock is required to ensure that only one coroutine can connect at a time to the same Modbus client.
        Also it prevents multiple attempts to connect to the same client
        while another coroutine is already trying to connect.

        For example, if two coroutines try to connect to the same Modbus client
        (address: 0.0.0.0:502, unitId: 1 and address: 0.0.0.0:502, unitId: 2)
        at the same time without lock the following error occurs:
        'Factory protocol connect callback called while connected.'
        """

        async with self.lock:
            if not self.__client.connected:
                await self.__client.connect()

    @with_lock_for_serial
    async def close(self):
        self.__client.close()

    @with_lock_for_serial
    async def read_coils(self, address, count, unit_id):
        result = await self.__client.read_coils(address=address, count=count, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def read_discrete_inputs(self, address, count, unit_id):
        result = await self.__client.read_discrete_inputs(address=address, count=count, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def read_holding_registers(self, address, count, unit_id):
        result = await self.__client.read_holding_registers(address=address, count=count, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def read_input_registers(self, address, count, unit_id):
        result = await self.__client.read_input_registers(address=address, count=count, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def write_coil(self, address, value, unit_id):
        result = await self.__client.write_coil(address=address, value=value, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def write_register(self, address, value, unit_id):
        result = await self.__client.write_register(address=address, value=value, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def write_coils(self, address, values, unit_id):
        result = await self.__client.write_coils(address=address, values=values, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    @with_lock_for_serial
    async def write_registers(self, address, values, unit_id):
        result = await self.__client.write_registers(address=address, values=values, slave=unit_id) # noqa
        self.__previous_request_time = int(monotonic() * 1000)
        return result

    def get_available_functions(self):
        return {
            1: self.read_coils,
            2: self.read_discrete_inputs,
            3: self.read_holding_registers,
            4: self.read_input_registers,
            5: self.write_coil,
            6: self.write_register,
            15: self.write_coils,
            16: self.write_registers,
        }

    @staticmethod
    def configure_master(config):
        framer = FramerType[config.method]

        if config.type == 'tcp' and config.tls:
            master = AsyncModbusTlsClient(config.host,
                                          config.port,
                                          framer,
                                          timeout=config.timeout,
                                          retries=config.retries,
                                          **config.tls)
        elif config.type == 'tcp':
            master = AsyncModbusTcpClient(host=config.host,
                                          port=config.port,
                                          framer=framer,
                                          timeout=config.timeout,
                                          retries=config.retries)
        elif config.type == 'udp':
            master = AsyncModbusUdpClient(host=config.host,
                                          port=config.port,
                                          framer=framer,
                                          timeout=config.timeout,
                                          retries=config.retries)
        elif config.type == 'serial':
            master = AsyncModbusSerialClient(port=config.port,
                                             timeout=config.timeout,
                                             retries=config.retries,
                                             baudrate=config.baudrate,
                                             stopbits=config.stopbits,
                                             bytesize=config.bytesize,
                                             parity=config.parity,
                                             handle_local_echo=config.handle_local_echo,
                                             framer=framer)
        else:
            raise Exception("Invalid Modbus transport type.")

        return master



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from os.path import dirname, getsize, exists
from sqlite3 import DatabaseError, ProgrammingError, InterfaceError, OperationalError
from time import sleep, monotonic, time
from logging import getLogger
from threading import Event, Thread, Lock
from queue import Queue, Empty
import datetime
from typing import Callable

from thingsboard_gateway.storage.sqlite.database_connector import DatabaseConnector
from thingsboard_gateway.storage.sqlite.storage_settings import StorageSettings


class Database(Thread):
    """
    Handles SQLite database operations for ThingsBoard Gateway:
    - Table creation & migration
    - Writing & reading messages efficiently
    - Deleting old records based on timestamp
    - Using PRIMARY KEY (`id`) for fast operations
    """

    def __init__(
            self,
            settings: StorageSettings,
            processing_queue: Queue,
            logger,
            stopped: Event,
            should_read: bool = True,
            should_write: bool = True,
            on_rotate_callback: Callable = None,
    ):
        self.__initialized = False
        self.__log = logger
        super().__init__()
        self.name = "DatabaseThread"
        self.daemon = True
        self.stopped = stopped
        self.database_stopped_event = Event()
        self._rotation_lock = Lock()
        self.__should_read = should_read
        self.__should_write = should_write
        self.__reached_size_limit = False
        self.settings = settings
        self._on_rotate_callback = on_rotate_callback
        self.directory = dirname(self.settings.data_file_path)
        self.db = DatabaseConnector(
            self.settings.data_file_path, self.__log, self.database_stopped_event
        )
        self.db.connect()
        self.init_table()
        self.process_queue = processing_queue
        self.__last_msg_check = 0
        self.__can_prepare_new_batch = True
        self.__next_batch = []
        self.__initialized = True

    def init_table(self):
        try:
            try:

                result = self.db.execute_read(
                    "SELECT sql FROM sqlite_master WHERE type='table' AND name='messages';"
                ).fetchone()

            except OperationalError as e:
                result = None
                self.__log.trace(
                    "Failed to execute read query in database:", exc_info=e
                )

            if result:
                if "timestamp" not in result[0] or "id" not in result[0]:
                    self.__log.info("Old schema detected! Starting migration...")
                    self.migrate_old_data()
            self.db.execute_write(
                """CREATE TABLE IF NOT EXISTS messages (
                                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                                        timestamp INTEGER NOT NULL,
                                        message TEXT NOT NULL
                                    );"""
            )
            cursor = self.db.execute_write(
                "CREATE INDEX IF NOT EXISTS idx_timestamp ON messages (timestamp);"
            )
            cursor.close()
            self.db.commit()

        except Exception as e:
            self.db.rollback()
            self.__log.exception("Failed to create table or migrate data! Error: %s", e)

    def run(self):
        self.__log.debug("Database thread started %r", id(self))
        interval = self.settings.oversize_check_period * 60
        sleep_time = 0.2

        last_time = monotonic()
        while not self.stopped.is_set() and not self.database_stopped_event.is_set():
            try:
                processing_started = monotonic()
                if self.__should_read:
                    if self.__can_prepare_new_batch and not self.__next_batch:
                        self.__next_batch = self.read_data()
                        self.__can_prepare_new_batch = False
                if self.__should_write:
                    self.process()

                remaining = sleep_time - (monotonic() - processing_started)
                if remaining > 0 and self.process_queue.empty():
                    sleep(remaining)
                if not self.__reached_size_limit:
                    now = monotonic()
                    if now - last_time >= interval:
                        last_time = now
                        self.process_file_limit()

            except Exception as e:
                self.__log.exception("Error in database thread: %s", exc_info=e)
        self.__log.debug("Database thread stopped %r", id(self))
        self.db.close()

    def process_file_limit(self):
        if exists(self.db.data_file_path):
            try:

                if getsize(self.db.data_file_path) >= float(self.settings.size_limit) * 1000000:
                    self.__reached_size_limit = True
                    if self._on_rotate_callback and self.__should_write:
                        with self._rotation_lock:
                            self._on_rotate_callback()
                    return True

            except FileNotFoundError as e:
                self.__reached_size_limit = True
                self.__log.debug("File is not found it is likely you deleted it ")
                self.__log.exception("Failed to find file ! Error: %s", e)

    def process(self):
        try:
            cur_time = int(time() * 1000)
            if (
                    cur_time - self.__last_msg_check
                    >= self.__last_msg_check + self.settings.messages_ttl_check_in_hours
                    and not self.stopped.is_set()
                    and not self.database_stopped_event.is_set()
            ):
                self.__last_msg_check = cur_time
                self.delete_data_lte(self.settings.messages_ttl_in_days)
            if not self.process_queue.empty():
                batch = []
                start_collecting = monotonic()
                while (
                        len(batch) < self.settings.batch_size
                        and not self.stopped.is_set()
                        and monotonic() - start_collecting < 0.1
                ):
                    try:
                        batch.append((cur_time, self.process_queue.get_nowait()))

                    except Empty:
                        if monotonic() - start_collecting > 0.1:
                            break
                        sleep(0.01)

                if batch:
                    start_writing = monotonic()

                    self.db.execute_many_write(
                        """INSERT INTO messages (timestamp, message) VALUES (?, ?);""",
                        batch,
                    )

                    self.db.commit()

                    self.__log.trace(
                        "Wrote %d records in %.2f ms, queue size: %d, Avg time per 1 record: %.2f ms",
                        len(batch),
                        (monotonic() - start_writing) * 1000,
                        self.process_queue.qsize(),
                        (monotonic() - start_writing) * 1000 / len(batch),
                    )
            else:
                self.database_stopped_event.wait(0.1)

        except Exception as e:
            self.db.rollback()
            self.__log.exception("Failed to write data to storage! Error: %s", e)

    def clean_next_batch(self):
        self.__next_batch = []

    def database_has_records(self) -> bool:
        """
        Returns True if there's at least one row in messages, False otherwise.
        """
        try:
            cursor = self.db.execute_read("SELECT EXISTS(SELECT 1 FROM messages);")
            if not cursor:
                return False
            row = cursor.fetchone()
            return bool(row[0])
        except (DatabaseError, ProgrammingError, InterfaceError) as e:
            self.__log.debug("Error checking for records: %s", e)
            return False
        except MemoryError:
            self.__log.debug("Out of memory checking for records")
            return False

    def read_data(self):
        if self.database_stopped_event.is_set() or not self.__initialized:
            return []
        try:
            if self.db.closed or self.stopped.is_set() or not self.db.connection:
                return []
            if self.__next_batch:
                return self.__next_batch
            start_time = monotonic()
            data = self.db.execute_read(
                """SELECT id, timestamp, message FROM messages LIMIT ?;""",
                (self.settings.max_read_records_count,),
            )
            if not data:
                return []
            collected_data = data.fetchall()
            elapsed_time = (monotonic() - start_time) * 1000
            if collected_data:
                self.__log.trace(
                    "Read %d records in %.2f ms", len(collected_data), elapsed_time
                )
            return collected_data
        except DatabaseError:
            return []
        except (ProgrammingError, InterfaceError) as e:
            self.__log.debug("Error reading data from storage: %s", e)
            return []
        except MemoryError:
            return []

    def interrupt(self):
        self.db.interrupt()

    def delete_data(self, row_id):
        if self.database_stopped_event.is_set():
            return
        try:
            data = self.db.execute_write(
                """DELETE FROM messages WHERE id <= ?;""",
                [
                    row_id,
                ],
            )
            self.db.commit()
            return data
        except Exception as e:
            self.db.rollback()
            self.__log.exception("Failed to delete data from storage! Error: %s", e)

    def delete_data_lte(self, days):
        if self.database_stopped_event.is_set():
            return
        try:
            ts = (datetime.datetime.now() - datetime.timedelta(days=days)).timestamp()
            data = self.db.execute_write(
                """DELETE FROM messages WHERE timestamp <= ? ;""", [ts]
            )
            self.db.commit()
            return data
        except Exception as e:
            self.db.rollback()
            self.__log.exception("Failed to delete data from storage! Error: %s", e)

    def migrate_old_data(self):
        if self.database_stopped_event.is_set():
            return
        try:
            self.__log.info("Renaming old table...")
            self.db.execute_write("ALTER TABLE messages RENAME TO messages_old;")
            self.db.commit()

            self.__log.info("Creating new optimized table...")
            self.db.execute_write(
                """CREATE TABLE messages (
                                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                                        timestamp INTEGER NOT NULL,
                                        message TEXT NOT NULL
                                    );"""
            )
            self.db.commit()

            self.__log.info("Migrating old data to new table...")
            self.db.execute_write(
                """
                INSERT INTO messages (timestamp, message)
                SELECT timestamp, message FROM messages_old;
            """
            )
            self.db.commit()

            self.__log.info("Dropping old table...")
            self.db.execute_write("DROP TABLE messages_old;")
            self.db.commit()

            self.__log.info("Migration completed successfully!")

        except Exception as e:
            self.db.rollback()
            self.__log.exception("Failed to migrate old data! Error: %s", e)

    def get_stored_messages_count(self) -> int:
        if self.database_stopped_event.is_set():
            return -1

        try:
            cursor = self.db.execute_read("SELECT COUNT(*) FROM messages;")
            if cursor is None:
                return -1

            row = cursor.fetchone()
            if not row:
                return 0

            return row[0]

        except (DatabaseError, InterfaceError) as e:
            self.__log.exception("Failed to query message count from SQLite: %s", e)
            return -1

    def close_db(self):
        if not self.database_stopped_event.is_set():
            self.database_stopped_event.set()

    def can_prepare_new_batch(self):
        self.__next_batch = []
        self.__can_prepare_new_batch = True
        return self.__can_prepare_new_batch

    def update_logger(self):
        self.__log = getLogger("storage")
        self.db.update_logger(logger=self.__log)
        self.__log.info("Logger updated")

    @property
    def should_read(self):
        return self.__should_read

    @property
    def should_write(self):
        return self.__should_write

    @property
    def reached_size_limit(self):
        return self.__reached_size_limit

    @should_read.setter
    def should_read(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("should_read must be a boolean")
        self.__should_read = value

    @should_write.setter
    def should_write(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("should_write must be a boolean")
        self.__should_write = value



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from time import monotonic
from threading import Thread, Event, Lock
from typing import Optional, Tuple, Dict

from thingsboard_gateway.gateway.constants import ReportStrategy, STRATEGIES_WITH_REPORT_PERIOD
from thingsboard_gateway.gateway.entities.datapoint_key import DatapointKey
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig


class ReportStrategyDataRecord:
    __slots__ = ["_value", "_device_name", "_device_type", "_connector_name",
                 "_connector_id", "_report_strategy", "_last_report_time", "_is_telemetry", "_ts"]

    def __init__(self, value, device_name, device_type, connector_name, connector_id, report_strategy, is_telemetry):
        self._value = value
        self._device_name = device_name
        self._device_type = device_type
        self._connector_name = connector_name
        self._connector_id = connector_id
        self._report_strategy: ReportStrategyConfig = report_strategy
        self._last_report_time = None
        self._is_telemetry = is_telemetry
        self._ts = None
        # TODO: Add aggregation functionality
        # self.__aggregated_data = {
        #     'sum': 0,
        #     'count': 0,
        #     'min': None,
        #     'max': None,
        #     'median': None,
        # }

    def get_value(self):
        return self._value

    def get_ts(self):
        return self._ts

    def is_telemetry(self):
        return self._is_telemetry

    def update_last_report_time(self, update_time):
        self._last_report_time = update_time

    def update_value(self, value):
        self._value = value

    def update_ts(self, ts):
        self._ts = ts

    def should_be_reported_by_period(self, current_time):
        if self._report_strategy.report_strategy in STRATEGIES_WITH_REPORT_PERIOD:
            return (self._last_report_time is None
                    or current_time - self._last_report_time + 50 >= self._report_strategy.report_period)
        else:
            return False

    def to_send_format(self):
        return (self._connector_name, self._connector_id, self._device_name, self._device_type), self._value

    @property
    def report_strategy(self):
        return self._report_strategy


class ReportStrategyDataCache:
    def __init__(self, config, logger):
        self._config = config
        self._data_cache: Dict[Tuple, Tuple[ReportStrategyDataRecord, float]] = {}
        self._lock = Lock()
        self._cleanup_interval = self._config.get("reportStrategyDataCacheCleanupInterval", 3600)
        self._stop_event = Event()
        self._cleanup_thread = Thread(target=self._cleanup_loop, daemon=True,
                                      name="Reporting strategy data cache cleanup thread")
        self._cleanup_thread.start()
        current_time = monotonic()
        self.__previous_cleanup_time = current_time
        self.__data_cache_current_ts = current_time
        self.__logger = logger

    def put(self, datapoint_key: DatapointKey, data: str, device_name,
            device_type, connector_name, connector_id, report_strategy,
            is_telemetry):
        key = (datapoint_key, device_name, connector_id)
        expire_ts = self.__data_cache_current_ts + report_strategy.ttl if report_strategy.ttl else 0
        record = ReportStrategyDataRecord(
            data, device_name, device_type, connector_name,
            connector_id, report_strategy, is_telemetry
        )
        with self._lock:
            self._data_cache[key] = (record, expire_ts)

    def get(self, datapoint_key: DatapointKey, device_name, connector_id) -> Optional[ReportStrategyDataRecord]:
        key = (datapoint_key, device_name, connector_id)
        with self._lock:
            item = self._data_cache.get(key)
            if not item:
                return None
            record, expire_ts = item
            if 0 < expire_ts < self.__data_cache_current_ts:
                del self._data_cache[key]
                return None
            return record

    def update_last_report_time(self, datapoint_key: DatapointKey, device_name, connector_id, update_time):
        record = self.get(datapoint_key, device_name, connector_id)
        if record:
            record.update_last_report_time(update_time)

    def update_key_value(self, datapoint_key: DatapointKey, device_name, connector_id, value):
        record = self.get(datapoint_key, device_name, connector_id)
        if record:
            record.update_value(value)
            with self._lock:
                expire_ts = self.__data_cache_current_ts + record.report_strategy.ttl if record.report_strategy.ttl else 0
                self._data_cache[(datapoint_key, device_name, connector_id)] = (record, expire_ts)

    def update_ts(self, datapoint_key: DatapointKey, device_name, connector_id, ts):
        record = self.get(datapoint_key, device_name, connector_id)
        if record:
            record.update_ts(ts)
            with self._lock:
                expire_ts = self.__data_cache_current_ts + record.report_strategy.ttl if record.report_strategy.ttl else 0
                self._data_cache[(datapoint_key, device_name, connector_id)] = (record, expire_ts)

    def delete_all_records_for_connector_by_connector_id(self, connector_id):
        with self._lock:
            keys_to_delete = [key for key in self._data_cache if key[2] == connector_id]
            for key in keys_to_delete:
                del self._data_cache[key]

    def clear(self):
        with self._lock:
            self._data_cache.clear()

    def stop(self):
        self._stop_event.set()
        self._cleanup_thread.join()

    def _cleanup_loop(self):
        while not self._stop_event.wait(1):
            self.__data_cache_current_ts = monotonic()
            if self.__previous_cleanup_time - self.__data_cache_current_ts >= self._cleanup_interval:
                self.__previous_cleanup_time = monotonic()
                with self._lock:

                    keys_to_delete = []
                    for key, (report_data_record, expire_ts) in self._data_cache.items():
                        if report_data_record.report_strategy.report_strategy != ReportStrategy.ON_RECEIVED and \
                                0 < expire_ts < self.__data_cache_current_ts:
                            keys_to_delete.append(key)
                    for key in keys_to_delete:
                        del self._data_cache[key]
                        self.__logger.debug("Removed expired record from cache: %s", key)
